#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: Ellipse.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




import math

# input : ((x, y), (a, b)) (origin is center of ellipse, a and b are the ellipse axis lengths)
# output: (angle, dist)
# angle is the angle of the normal to the ellipse (in degrees) (The latitude in the original paper)
# dist is the distance from the the point to the ellipse
#
# algorithm from:
#    K.M. BORKOWSKI
#    ACCURATE ALGORITHMS TO TRANSFORM GEOCENTRIC TO GEODETIC COORDINATES
#    http://www.astro.uni.torun.pl/~kb/Papers/geod/Geod-BG.htm
#    
#    based on code found on the web at http://www.blastnet.ee/etm/temp/geo.py
#    That python code is too similar to the original fortran to have its own copyright

def distToEllipse((x, y), (a, b)):
    angle = 0.0
    dist = 0.0

    if x == 0.0:
        if y < 0.0:
            return (-90.0, -y - b)
        else:
            return (90.0, y - b)

    angleXswap = (x < 0.0)
    x = abs(x)

    if y < 0.0:
        b = -b
    E = ((y + b) * b / a - a) / x
    F = ((y - b) * b / a + a) / x
    P = (E * F + 1.0) * 4.0 / 3.0
    Q = (E * E - F * F) * 2.0
    D = P * P * P + Q * Q

    # print "E: " , E , " F: " , F, " P: " , P , " Q: " , Q , " D: " , D

    if (D >= 0.0):
        ss = math.sqrt(D) + Q
        s = math.exp( math.log( abs(ss) ) / 3.0 )
        if ss < 0.0:
            s = -s
        v = P / s - s
    else:
        v = 2.0 * math.sqrt(-P) * math.cos( math.acos(Q / P / math.sqrt(-P)) / 3.0)

    G = 0.5 * (E + math.sqrt(E * E + v))
    t = math.sqrt(G * G + (F - v * G) / (G + G - E)) - G

    # extract the denominator of the final calculation - may go to zero if you're near x == 0.0
    denom = (2.0 * b * t)
    if denom == 0.0:
        if y < 0.0:
            return (-90.0, -y + b) # note that b had it's sign reversed above
        else:
            return (90.0, y - b)
    else: 
        angle = math.atan((1.0 - t * t) * a / denom)

    if angleXswap:
        angle = math.pi - angle

    dist = (x - a * t) * math.cos(angle) + (y - b) * math.sin(angle)
    angle = math.degrees( angle )

    return (angle, dist)

# given a normal angle from a point on an ellipse, return the parameter t for that
# point in the parameterisation (a * cos(t), b * sin(t))
def getReducedAngle(angle, (a, b)):
    radAngle = math.radians(angle)
    otherQuadrant = (abs(radAngle) > (math.pi/2.0))
    
    tn = math.tan(radAngle)
    radReduceAngle = math.atan(b/a*tn)
    if otherQuadrant:
        radReduceAngle += math.pi
    return math.degrees(radReduceAngle)

# given a reduced angle to a point on an ellipse (see getReducedAngle())
# return the normal angle from that point
def getNormalAngle(angle, (a, b)):
    radAngle = math.radians(angle)
    otherQuadrant = (abs(radAngle) > (math.pi/2.0))
    
    tn = math.tan(radAngle)
    radReduceAngle = math.atan(a/b*tn)
    if otherQuadrant:
        radReduceAngle += math.pi
    return math.degrees(radReduceAngle)
