#
#   Copyright 2004 The University of New South Wales (UNSW) and National  
#   ICT Australia (NICTA).
#
#   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
#   redistribute it and/or modify it under the terms of the GNU General  
#   Public License as published by the Free Software Foundation; either  
#   version 2 of the License, or (at your option) any later version as  
#   modified below.  As the original licensors, we add the following  
#   conditions to that license:
#
#   In paragraph 2.b), the phrase "distribute or publish" should be  
#   interpreted to include entry into a competition, and hence the source  
#   of any derived work entered into a competition must be made available  
#   to all parties involved in that competition under the terms of this  
#   license.
#
#   In addition, if the authors of a derived work publish any conference  
#   proceedings, journal articles or other academic papers describing that  
#   derived work, then appropriate academic citations to the original work  
#   must be included in that publication.
#
#   This rUNSWift source is distributed in the hope that it will be useful,  
#   but WITHOUT ANY WARRANTY; without even the implied warranty of  
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along  
#   with this source code; if not, write to the Free Software Foundation,  
#   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# 
#  Last modification background information
#  $Id: LocChallenge.py 4365 2004-09-22 07:14:44Z kcph007 $
# 
#  Copyright (c) 2004 UNSW
#  All Rights Reserved.
# 




#===============================================================================
#   Python Behaviours : 2004 (c) 
#
# LocChallenge.py
#
# The localisation challenge behaviour. Moves to a series of points on the
#   field, and wags its tail at each point. Uses some stuff in VisualCortex
#   and GPS that is specific to the challenge, which you should turn on with
#   the flag_LOC_CHALLENGE variable in share/Common.h.
#
#===============================================================================

"""
Localisation challenger 2004
"""
import Constant
import Global
import VisionLink
import ActiveLocalise
import HelpShort
import HelpLong
import HelpTrack
import math

#===============================================================================
# Global variables of this module.
#===============================================================================

POS_TIME = 7                    # Number of frames to confirm location
POS_ACCURACY = 3                # Position accuracy required to proceed
POS_MAX_VARIANCE = 10000        # Max position variance allowed to proceed
SURVEY_TIME = 1500              # How many frames dog should survey for
ORIENTATE_TIME = 300            # How many frames dog should stat. localise for
WATCH_FEATURE = 40              # How long dog will focus on an edge feature

# Global coordinates of field line features to face when close to destination
FACING_POSITIONS = [
    (Constant.LEFT_GOALBOX_EDGE_X, Constant.LEFT_GOALBOX_EDGE_Y),
    (Constant.RIGHT_GOALBOX_EDGE_X, Constant.RIGHT_GOALBOX_EDGE_Y),
    (Constant.LEFT_GOALBOX_EDGE_X, Constant.OWN_GOALBOX_EDGE_Y),
    (Constant.RIGHT_GOALBOX_EDGE_X, Constant.OWN_GOALBOX_EDGE_Y),
    (Constant.WALL_THICKNESS, Constant.FIELD_LENGTH / 2),
    (Constant.FIELD_WIDTH - Constant.WALL_THICKNESS, Constant.FIELD_LENGTH / 2)
]

# Global coordinates of field line features to localise off
LOCALISE_POSITIONS = FACING_POSITIONS + [
    (Constant.FIELD_WIDTH / 2, Constant.FIELD_LENGTH / 2),
    (Constant.WALL_THICKNESS, 80),
    (Constant.WALL_THICKNESS, 360),
    (Constant.FIELD_WIDTH - Constant.WALL_THICKNESS, 80),
    (Constant.FIELD_WIDTH - Constant.WALL_THICKNESS, 360)
]

points = []                     # These are the points the dog should move to
pointSequence = []              # The order of visiting the points
moving = True                   # Flag to indicate if dog is moving
faceDest = True                 # Should dog face destination as it moves
turning = True                  # Whether dog should only turn
paused = False                  # Is the dog paused or not
initialised = False             # Has the points file been loaded yet?
pathSelected = False            # Have we chosen the point sequence to visit?
facingPoint = (0, 0)            # Which line feature to face when near dest.
movePoint = (0, 0)              # Coordinates of destination point to move to
moveToPoint = 0                 # Seq index of the point the dog is moving to
distanceToPoint = 1000          # How far dog is from destination point 
localiseCounter = 0             # Determines when to switch localising features
positionedCounter = 0           # How long the dog has been at the point
wagTailCounter = 0              # Keeps track of tail wagging progress
surveyCounter = 0               # How long surveying has been going on
orientateCounter = 0            # How long re-orientation has been going
horizonPan = 0                  # When looking up, this is the pan value to use

useMinimalMode = True           # This makes the dog skip initial stationary
                                # localise stage when first turned on


#===============================================================================
# Functions of this module.
#===============================================================================
# The top level decisions for this player

def DecideNextAction():
    global surveyCounter, wagTailCounter, positionedCounter, orientateCounter
    global initialised, paused, moving

    Global.frameReset()
    HelpLong.setDefaultAction()
    HelpLong.setDefaultIndicators()
    
    # Turn face lights on if we're seeing pink blobs to localise off
    if VisionLink.getUsePinkUpdate() == 1:
        HelpLong.showFacePattern((2, 1, 0, 1, 2))
    else:
        HelpLong.showFacePattern((0, 0, 0, 0, 0))

    # Load the points file
    if not initialised:
        InitialiseChallenge()

    if paused:
        DoPause()
        return
    
    # If we're in the mapping stage of the challenge, continue doing the survey
    if surveyCounter < SURVEY_TIME:
        DoSurvey()
        return
    
    # If we got this far, we're in the second stage of the challenge, and so are
    # moving to each of the five points loaded from the points file
    
    # Re-orientate (we might have been moved during the break)
    if orientateCounter <= ORIENTATE_TIME:
        if orientateCounter == 0:
            VisionLink.resetEdgeMap()
        
        headPan = (orientateCounter % 120) * 3 - 90.0
        if headPan > 90:
            headTilt = -30
            headPan = 180 - headPan
        else:
            headTilt = -5

        HelpLong.setHeadParams(headPan, headTilt, 0,Constant.HTAbs_h)

        #if orientateCounter == ORIENTATE_TIME:
        #    VisionLink.setGlobalMaxEdge()

        orientateCounter = orientateCounter + 1            
        return

    # Find the best path between points
    if not pathSelected:
        FindShortestPath()    

    # Calculate gps coordinates of the destination point we want to move to
    CalcMovePoint()
    
    # Check if we've arrived at destination
    if AtDestPoint():
        positionedCounter = positionedCounter + 1
    else:
        positionedCounter = 0
    
    # If we've been at the destination long enough, stop moving
    if positionedCounter > POS_TIME:
        moving = False
        
    if moving:
        Localise()
        MoveToPoint()
    else:
        WagTail()
        
        if wagTailCounter > 180:
            SwitchToNextPoint()
        
#-------------------------------------------------------------------------------
# Loads the points file

def InitialiseChallenge():
    global points, pointSequence, initialised

    points = []

    try:
        # Parse the coordinates contained in the points file
        f = open("/MS/points.cfg")
        for line in f.readlines():
            # Load a single coordinate pair
            centreCoord = map(lambda x:int(x), line.split())

            if len(centreCoord) == 2:
                # Convert challenge coordinates to gps coordinates
                if VisionLink.getTeamColor() == 1:
                    # Convert to red team coordinates
                    gpsCoord = (Constant.FIELD_WIDTH / 2 - centreCoord[1],
                                centreCoord[0] + Constant.FIELD_LENGTH / 2)
                else:
                    # Convert to blue team coordinates
                    gpsCoord = (centreCoord[1] + Constant.FIELD_WIDTH / 2,
                                Constant.FIELD_LENGTH / 2 - centreCoord[0])
        
                # Add to points list
                points = points + [gpsCoord]
    
        f.close()

    except:
        points = []
        
    # If file wasn't loaded, use a default set of points (for testing only!)
    if len(points) < 5:
        print "Coordinate file not loaded successfully, using default points"
        points = [(30,380), (95,60), (200,410), (70,220), (220,70)]

    initialised = True

#-------------------------------------------------------------------------------
# Waits until head is pressed

def DoPause():
    global paused

    for pressSensor in range(1): # Used to react to back sensors
        if VisionLink.getPressSensorCount(pressSensor) > 8:
            paused = False;
            VisionLink.resetEdgeMap()
            VisionLink.setIfGPSMotionUpdate(True)
            VisionLink.setIfGPSVisionUpdate(True)
            HelpLong.setDefaultIndicators()

#-------------------------------------------------------------------------------
# Scans the field borders for a duration close to a minute

def DoSurvey():
    global surveyCounter, paused, faceDest

    surveyCounter = surveyCounter + 1

    HelpLong.setNormalWalk(0, 0, 0)
    HelpLong.setDefaultAction()
    VisionLink.setAllowPinkMapping(False)
    selfHead = Global.selfLoc.getHeading()
    
    # Initial localisation to get orientated
    if surveyCounter < 150:
        faceDest = 0
        HelpTrack.stationaryLocalise()
    
    # Now get into starting position for pan scan
    elif surveyCounter < 300:
        if ((surveyCounter - 150) % 30) == 0:
            ActiveLocalise.AlternateSetBeacon()
        
        ActiveLocalise.DecideNextAction()
        turnCCW = HelpShort.normalizeAngle_180(faceDest - selfHead)
        HelpLong.setNormalWalk(0, 0, HelpShort.CLIP(turnCCW, 40))

    # Localise a bit more
    elif surveyCounter < 400:
        if ((surveyCounter - 300) % 25) == 0:
            ActiveLocalise.SmartSetBeacon()
            
        ActiveLocalise.DecideNextAction()
    
    # Move head to start of pan scan
    elif surveyCounter < 420:
        HelpLong.setHeadParams(-90, 0, 0, Constant.HTAbs_h)

    # Do a 180 degree head pan at each quadrant direction
    elif surveyCounter < 1260:
        scanCount = (surveyCounter - 420) % 220

        # Do a slow head pan
        if scanCount < 180:
            VisionLink.setAllowPinkMapping(True)
            HelpLong.setHeadParams(scanCount - 90, 0, 0, Constant.HTAbs_h)

        # Select next quadrant
        elif scanCount == 180:
            faceDest = HelpShort.normalizeAngle_0_360(faceDest + 90)
            ChooseLeftmostBeacon()

        # Then turn 90 degrees            
        else:
            turnCCW = HelpShort.normalizeAngle_180((faceDest - selfHead) * 0.5)
            HelpLong.setNormalWalk(0, 0, HelpShort.CLIP(abs(turnCCW), 30))
            ActiveLocalise.DecideNextAction()
    
    # Move to centre position    
    elif surveyCounter < SURVEY_TIME - 1:
        HelpLong.setNormalWalk(0, 0, 0)
        HelpTrack.saGoToTargetFacingHeading(145, 220, 270, 3, 10)
        HelpTrack.stationaryLocalise()

    else:
        paused = True
        VisionLink.setIfGPSVisionUpdate(False)
        VisionLink.setUseBeacons(False)

#-------------------------------------------------------------------------------
# Makes the head look at points of interest (field line points to localise off)

def Localise():
    global localiseCounter, horizonPan

    selfLoc = Global.selfLoc
    selfHead = selfLoc.getHeading()
    possibleFeatures = []
    
    # Find all line features within the robot's field of view
    for featureLoc in LOCALISE_POSITIONS:
        relPos = map(lambda b,s:b-s, featureLoc, selfLoc.getPos())
        featureDist = HelpShort.getLength(relPos)
        if featureDist == 0:
            featureHead = selfHead
        else:
            featureHead = HelpShort.RAD2DEG(math.atan2(relPos[1], relPos[0]))
            
        localHead = HelpShort.normalizeAngle_180(featureHead - selfHead)
        
        if abs(localHead) < 80 and featureDist > 15 and featureDist < 200:
            possibleFeatures.append(tuple(relPos))

    # Use localiseCounter to determine which feature to look at
    localiseCounter = localiseCounter + 1
    if localiseCounter >= len(possibleFeatures) * WATCH_FEATURE:
        localiseCounter = -15

        horizonPan = horizonPan + 30
        if horizonPan > 30:
            horizonPan = -30

    # Look up every once in a while in the hope of seeing a pink blob.            
    if localiseCounter < 0:
        HelpLong.setHeadParams(horizonPan, 0, 0, Constant.HTAbs_h)
        
    # Otherwise look at an edge feature
    else:
        useFeature = localiseCounter / WATCH_FEATURE
        relPos = HelpShort.rotate(possibleFeatures[useFeature], 90 - selfHead)
        targetX = -relPos[0]
        targetY = relPos[1]

        HelpLong.setHeadParams(targetX, 0, targetY, Constant.HTAbs_xyz)

#-------------------------------------------------------------------------------
# Calculates the gps coordinates that the dog should move to

def CalcMovePoint():
    global points, pointSequence, moveToPoint, movePoint
    global facingPoint, faceDest        

    point = points[pointSequence[moveToPoint]]
    facingVec = (facingPoint[0] - point[0], facingPoint[1] - point[1])
    facingVec = HelpShort.normalise(facingVec)
    
    if faceDest:
        # If far away from destination point, move to a point behind the
        # actual destination, so we when we get near, we can move forward
        # towards our selected line feature, and localise better.
        facingVec = (-5 * facingVec[0], -5 * facingVec[1])
    
    else:
        # If we are close, move to the point that would position the centre
        # of the dogs body over the destination.
        facingVec = (10 * facingVec[0], 10 * facingVec[1])

    moveTo = [point[0] + facingVec[0], point[1] + facingVec[1]]
    movePoint = ClipToField(moveTo, 15)
        
#-------------------------------------------------------------------------------
# Returns true if the dog is close to the destination point, and has small
# enough self variance

def AtDestPoint():
    global distanceToPoint, movePoint

    pos = Global.selfLoc.getPos()
    posVar = Global.selfLoc.getPosVar()
    distanceToPoint = HelpShort.getLength(map(lambda m,p:m-p, movePoint, pos))
    
    return (distanceToPoint < POS_ACCURACY and posVar < POS_MAX_VARIANCE)

#-------------------------------------------------------------------------------
# Makes the dog move to the given point and heading

def MoveToPoint():
    global distanceToPoint, movePoint, facingPoint, turning, faceDest

    if distanceToPoint < 5:
        if faceDest:
            print "Stopped sprinting"
            
        faceDest = False

    selfLoc = Global.selfLoc.getPos()
    selfHead = Global.selfLoc.getHeading()

    # Choose a heading vector.
    if faceDest:
        # When far from location, head straight to it
        dX, dY = (movePoint[0] - selfLoc[0], movePoint[1] - selfLoc[1])
    else:
        # Face line feature (for localising) when close to destination
        dX, dY = (facingPoint[0] - selfLoc[0], facingPoint[1] - selfLoc[1])

    # Calculate the angle between our self angle and the heading vector (dH).
    if dX == 0 and dY == 0:
        head = selfHead
    else:
        head = HelpShort.RAD2DEG(math.atan2(dY, dX))
        head = HelpShort.normalizeAngle_0_360(head)

    dH = HelpShort.normalizeAngle_180(head - selfHead)

    # If far from destination, we want to sprint straight towards destination.
    if faceDest:
        if abs(dH) < 5:
            # If we are facing in the right direction, move forward
            turning = False
        if abs(dH) > 30:
            # If we are facing in the wrong direction, only turn
            turning = True
            
        if not turning:
            # When facing correct direction, sprint using elliptical walk
            HelpLong.setEllipticalWalk(0, 0, 0)
            HelpTrack.saGoToTarget(movePoint[0], movePoint[1], 8, 5)
        else:
            HelpLong.setNormalWalk(0, 0, HelpShort.CLIP(dH * 1.5, 40))

    # If close to destination point, we want to face the pre-determined line
    # feature.
    else:
        if abs(dH) < 5:
            turning = False
            
        if abs(dH) > 30:
            turning = True

        if turning:
            HelpLong.setNormalWalk(0, 0, HelpShort.CLIP(dH * 1.5, 40))
            
        elif distanceToPoint > POS_ACCURACY:
            # Reduce speeds as we approach the destination point
            speedLim = 1 + distanceToPoint / 5
            turnLim = 5 + abs(dH) / 6
        
            HelpLong.setNormalWalk(0, 0, 0)
            HelpTrack.saGoToTargetFacingHeading(movePoint[0], movePoint[1],
                                               head, speedLim, turnLim)
        else:
            HelpLong.setDefaultAction()

#-------------------------------------------------------------------------------
# Makes the dog wag its tail

def WagTail():
    global wagTailCounter
    
    wagTailCounter = wagTailCounter + 1

    if (wagTailCounter % 20) < 10:
        tailPos = Constant.IND_TAIL_RIGHT / 2
    else:
        tailPos = Constant.IND_TAIL_LEFT / 2

    Global.finalIndicators[Constant.iTailH] = tailPos
    Global.finalIndicators[Constant.iTailV] = Constant.IND_TAIL_DOWN

#-------------------------------------------------------------------------------
# Switches to the next point in the sequence

def SwitchToNextPoint():
    global paused, moving, wagTailCounter, positionedCounter
    global moveToPoint, points, pointSequence, turning, faceDest
    
    paused = True
    moving = True
    wagTailCounter = 0
    positionedCounter = 0
    VisionLink.setIfGPSMotionUpdate(False)
    VisionLink.setIfGPSVisionUpdate(False)

    # Switch to next point, cycle to first if necessary
    turning = True
    faceDest = True
    moveToPoint = moveToPoint + 1
    if moveToPoint >= len(points):
        moveToPoint = 0

    ChooseFacingPoint()
    
    print "Now moving to point ", points[pointSequence[moveToPoint]]

#-------------------------------------------------------------------------------
# Finds the shortest path between the five points to go to

def FindShortestPath():
    global points, pointSequence, pathSelected

    # Calculate distance look-up table
    num = len(points)
    dist = []
    allPoints = points + [Global.selfLoc.getPos()]
    aPRange = range(len(allPoints))

    for a in aPRange:
        dist.append([])
    
        for b in aPRange:
            delta = map(lambda x,y:x-y, allPoints[a], allPoints[b])
            dist[a].append(HelpShort.getLength(delta))
                           
    # Use backtracking to find shortest path between all points    
    (pointSequence, l) = GetShortestPath(num, range(num), dist, 1000 * num)

    ChooseFacingPoint()

    print "Points: ", points
    print "Best sequence: ", pointSequence
    
    pathSelected = True

#-------------------------------------------------------------------------------
# Recursively search all possibile point sequences for shortest path. If there
# were lots of points to visit, this wouldn't be used because it's too slow.
# But seeing there are only 5 points, we can afford to look for the optimal
# solution.

def GetShortestPath(prev, available, dist, max):
    l = 0
    best = max
    path = []
    check = []

    for choice in available:
        d = dist[prev][choice] 

        if d < best:
            newAvailable = available[:]
            newAvailable.remove(choice)
            (check, l) = GetShortestPath(choice, newAvailable, dist, best - d)
            l = d + l
            check = [choice] + check

            if len(check) == len(available):
                best = l
                path = check[:]

    if path == []:
        return ([], 0)
    else:
        return (path, best)
                
#-------------------------------------------------------------------------------
# Choose closest line feature to face towards (when dog gets close to
# destination).

def ChooseFacingPoint():
    global moveToPoint, facingPoint, points, pointSequence

    closestDist = Constant.FIELD_DIAGONAL
    for edgeLoc in FACING_POSITIONS:
        # Calculate distance from destination point to line feature
        p = points[pointSequence[moveToPoint]]
        delta = map(lambda e,p:e-p, edgeLoc, p)
        dist = HelpShort.getLength(delta)
        
        # We want to pick the line feature that is closest to the destination,
        # but not so close that it will be under the dog when it gets there
        if dist > 15 and dist < closestDist:
            closestDist = dist
            facingPoint = edgeLoc

#-------------------------------------------------------------------------------
# Selects the leftmost beacon to localise off, so that when we turn left we can
# continually track it.

def ChooseLeftmostBeacon():
    selfPos = Global.selfLoc.getPos()
    selfHead = Global.selfLoc.getHeading() 
    beaconPos = VisionLink.getGPSCoordArray()
    curCompHead = -100
    ActiveLocalise.localiseBeacon = 0    
    
    for beacon in ActiveLocalise.CHECK_BEACONS:
        relPos = map(lambda b,s:b-s, beaconPos[beacon], selfPos)
        if relPos[0] == 0 and relPos[1] == 0:
            beaconHead = 0
        else:
            beaconHead = HelpShort.RAD2DEG(math.atan2(relPos[1], relPos[0]))
        localHead  = HelpShort.normalizeAngle_180(beaconHead - selfHead)
    
        if abs(localHead) < 90 and localHead > curCompHead:
            curCompHead  = localHead
            ActiveLocalise.localiseBeacon = beacon


#-------------------------------------------------------------------------------
# Clips the given point to within the field by a buffer amount given by 'space'

def ClipToField(clipPoint, space):
    point = list(clipPoint)

    # Work in one quadrant, so there aren't as many checks
    flipX = (point[0] > Constant.FIELD_WIDTH / 2)
    flipY = (point[1] > Constant.FIELD_LENGTH / 2)
    
    if flipX:
        point[0] = Constant.FIELD_WIDTH - point[0]
    if flipY:
        point[1] = Constant.FIELD_LENGTH - point[1]

    if point[0] <= Constant.LEFT_GOAL_POST:
        # Clip to straight wall near goal.
        point[1] = max(point[1], space + Constant.WALL_THICKNESS)

    else:
        # Clip to the rear of the goal.
        point[1] = max(point[1], space - Constant.GOAL_LENGTH)
        
        if point[1] > Constant.WALL_THICKNESS:
            # Clip to corners of goal mouth.
            deltaGoalCorner = [0, 0]
            deltaGoalCorner[0] = point[0] - Constant.LEFT_GOAL_POST
            deltaGoalCorner[1] = point[1] - Constant.WALL_THICKNESS
            dist = HelpShort.getLength(deltaGoalCorner)

            if dist < space:
                scaler = (space - dist) / dist
                point[0] = point[0] + deltaGoalCorner[0] * scaler
                point[1] = point[1] + deltaGoalCorner[1] * scaler
                
        else:
            # Clip to goal sides.
            point[0] = max(point[0], space + Constant.LEFT_GOAL_POST)
        
    # Clip to the long straight wall sections.
    if point[1] > Constant.WALL_THICKNESS:
        point[0] = max(point[0], space + Constant.WALL_THICKNESS)
        
    # Clip to corners as well.
    wallWidth = Constant.WALL_THICKNESS 
    cornerConst = Constant.CORNER_WIDTH + (1.0 / math.sqrt(2) - 1.0) * wallWidth
    corner = (1, 1, -cornerConst)
    
    normaliseConst = 1.0 / HelpShort.getLength(corner[0:2])
    dist = corner[0] * point[0] + corner[1] * point[1] + corner[2]
    dist = dist * normaliseConst
        
    if dist < space:
        scaler = (space - dist) * normaliseConst
        point[0] = point[0] + corner[0] * scaler
        point[1] = point[1] + corner[1] * scaler

    # Make sure we don't move into overlapping buffer regions in
    # different quadrants.
    point[0] = min(point[0], Constant.FIELD_WIDTH / 2)
    point[1] = min(point[1], Constant.FIELD_LENGTH / 2)

    # Flip back to original quadrant
    if flipX:
        point[0] = Constant.FIELD_WIDTH - point[0]
    if flipY:
        point[1] = Constant.FIELD_LENGTH - point[1]

    return tuple(point)
