/*

   Copyright 2003 The University of New South Wales (UNSW) and National  
   ICT Australia (NICTA).

   This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
   redistribute it and/or modify it under the terms of the GNU General  
   Public License as published by the Free Software Foundation; either  
   version 2 of the License, or (at your option) any later version as  
   modified below.  As the original licensors, we add the following  
   conditions to that license:

   In paragraph 2.b), the phrase "distribute or publish" should be  
   interpreted to include entry into a competition, and hence the source  
   of any derived work entered into a competition must be made available  
   to all parties involved in that competition under the terms of this  
   license.

   In addition, if the authors of a derived work publish any conference  
   proceedings, journal articles or other academic papers describing that  
   derived work, then appropriate academic citations to the original work  
   must be included in that publication.

   This rUNSWift source is distributed in the hope that it will be useful,  
   but WITHOUT ANY WARRANTY; without even the implied warranty of  
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
   General Public License for more details.

   You should have received a copy of the GNU General Public License along  
   with this source code; if not, write to the Free Software Foundation,  
   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


/*
 * Last modification background information
 * $$
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 */

#include "KI2DWithVelocity.h"

// maximum number of frames between two position observations constitute a
// velocity observation
static const int MAX_FRAME = 4;

// rate at which deceleration is taken
static const double DECELERATION = 0.98;

static const double IVSumDecayRate = 0.4;


KI2DWithVelocity::KI2DWithVelocity(GPS *g,
		double iniX,
		double iniY,
		double iniXVar,
		double iniYVar,
		double iniVX,
		double iniVY,
		double iniVXVar,
		double iniVYVar) : KalmanInfo2D(g,
			iniX,
			iniY,
			iniXVar,
			iniYVar)
{
	velVal(0, 0) = iniVX;
	velVal(1, 0) = iniVY;
	velNorm2Global();
	velNorm2Local();

	velCov(0, 0) = iniVXVar;
	velCov(0, 1) = 0.0;
	velCov(1, 0) = 0.0;
	velCov(1, 1) = iniVYVar;

	lastBallMethod = 0;
	lastDist = 0.0;
	lastHead = 0.0;
	lastPos(0, 0) = 0.0;
	lastPos(1, 0) = 0.0;
	lastPos(2, 0) = 0.0;
	numFrames = MAX_FRAME;

	ivLength = 0.0;
}

const Vector & KI2DWithVelocity::getVelocity(int context) {
	if (context == GLOBAL) {
		if (!velGlobalCur)
			velNorm2Global();

		return gVelVal;
	}
	else {
		// context == LOCAL

		if (!velLocalCur)
			velNorm2Local();

		return lVelVal;
	}
}

void KI2DWithVelocity::infoKalman(const SlimVisOb &vob, int lastBallMeth) {
	KalmanInfo2D::infoKalman(vob);

	// position of robot
	const WMObj &self = gps->self();

	if (numFrames <= 0) {
		cout << "Error! : number of frames should be greater than zero ("
			<< numFrames << ")" << endl;
	}
	else if (numFrames <= MAX_FRAME) {
		// this and previous observation are close enough to create a velocity observation

		// the covariance of the position of the robot
		const MMatrix3 &rcov = gps->selfCov();

		// determine global position of last observation
		MVec2 lastObs;
		MMatrix2 lastObsCov;
		//if((gps->vbPull(0,0)+gps->vbPull(1,0)+gps->vbPull(2,0)) > 10) {
		//    cout << "pull = {" << gps->vbPull(0,0) << "," << gps->vbPull(1,0);
		//    cout << "," << gps->vbPull(2,0) << "}" << endl;
		//}
		
		(gps->vbPull) (0, 0) = 0.0;
		(gps->vbPull) (1, 0) = 0.0;
		(gps->vbPull) (2, 0) = 0.0;
		getGlobalObservation(lastPos,
				rcov,
				lastDist,
				lastHead,
				lastDistVar,
				lastHeadVar,
				lastObs,
				lastObsCov);

		// determine global position of this observation
		MVec2 curObs;
		MMatrix2 curObsCov;
		MVec3 curPos;
		curPos(0, 0) = self.pos.x;
		curPos(1, 0) = self.pos.y;
		curPos(2, 0) = self.h;
		getGlobalObservation(curPos,
				rcov,
				vob.d,
				vob.h,
				vob.var,
				vob.angleVar,
				curObs,
				curObsCov);

		// measurement of velocity is the difference betweeen the position observations (divided by numFrames)
		// covariance is the sum of the two
		MVec2 meas = curObs;
		meas -= lastObs;
        meas /= numFrames;

		if (lastBallMeth != lastBallMethod) {
			// the methods for finding distance to the ball are not the same so velocity measurement will be dodgy
		}
		else if (sqrt(SQUARE(meas(0, 0)) + SQUARE(meas(1, 0))) > 10.0) {
			// do nothing
		}
		else if (getNumSDVel(meas(0, 0), meas(1, 0)) > 1.0) {
			velCov *= 2.0;
		}
        else if (gps->motionForward + gps->motionLeft + gps->motionTurn > 0.0) {
            // Velocity estimation is bad when the dog is moving
		} else {
			MMatrix2 cov = curObsCov;
			cov += lastObsCov;

			cov ^= numFrames;

			// perform kalman filtering
			MVec2 iv = meas;
			iv -= velVal;

			double ivl = sqrt(SQUARE(iv(0, 0)) + SQUARE(iv(1, 0)));
			ivLength *= 1.0 - IVSumDecayRate;
			ivLength += IVSumDecayRate * ivl;

			MMatrix2 m1;
			m1 = velCov;
			m1 += cov;
			MMatrix2 m1Inv;
			m1Inv.isInverse(m1);
			MMatrix2 k;
			k.isMult(velCov, m1Inv);

			MVec2 diff;
			diff.isMult(k, iv);
			velVal += diff;

			MMatrix2 m3(1);
			m3 -= k;
			velCov.preMult(m3);

			velGlobalCur = false;
			velLocalCur = false;
		}
	}

	lastBallMethod = lastBallMeth;
	lastDist = vob.d;
	lastHead = vob.h;
	lastDistVar = vob.var;
	lastHeadVar = vob.angleVar;
	lastPos(0, 0) = self.pos.x;
	lastPos(1, 0) = self.pos.y;
	lastPos(2, 0) = self.h;
	numFrames = 0;

#ifndef PRE_SANITY_BLOB
	//  cout << "velocity={" << velVal(0,0) << "," << velVal(1,0) << "}" << endl;
#endif
}


void KI2DWithVelocity::growVariance(double amount, double velAmount) {
	KalmanInfo2D::growVariance(amount, velVal(0, 0), velVal(1, 0));

	MMatrix2 velMult(0);
	velMult(0, 0) = DECELERATION;
	velMult(1, 1) = DECELERATION;

	velVal.preMult(velMult);
	velCov.preMult(velMult);

	velCov(0, 0) += velAmount;
	velCov(1, 1) += velAmount;

	// if the ball has gone off the field then set velocity to zero
	if ((val(0, 0) < 0)
			|| (val(0, 0) > FIELD_WIDTH)
			|| (val(1, 0) < 0)
			|| (val(1, 0) > FIELD_LENGTH)) {
		velVal(0, 0) = 0.0;
		velVal(1, 0) = 0.0;
	}

	numFrames++;

	velGlobalCur = false;
	velLocalCur = false;
}


const MMatrix2 & KI2DWithVelocity::getVelCov() {
	return velCov;
}

MMatrix2 KI2DWithVelocity::getVCovLocal() {
	// get the covariance matrix in terms of local coordinates
	// angle to get the coords wrt robot to global coords
	double angleRad = DEG2RAD(gps->self().h - 90);
	double cs = cos(angleRad);
	double sn = sin(angleRad);
	MMatrix2 rotate(0);
	rotate(0, 0) = cs;
	rotate(0, 1) = -sn;
	rotate(1, 0) = sn;
	rotate(0, 1) = cs;
	MMatrix2 rotateInv(0);
	rotateInv(0, 0) = cs;
	rotateInv(0, 1) = sn;
	rotateInv(1, 0) = -sn;
	rotateInv(0, 1) = cs;

	MMatrix2 covWRTB;
	covWRTB.isMult(velCov, rotate);
	covWRTB.preMult(rotateInv);

	return covWRTB;
}


MMatrix2 KI2DWithVelocity::getVCovBall() {
	// get the covariance matrix in terms of the ball, y axis towards the robot
	// angle to get the coords wrt ball to global coords
	double angleRad = atan2(velVal(0, 0) - gps->self().pos.x,
			velVal(1, 0) - gps->self().pos.y);
	double cs = cos(angleRad);
	double sn = sin(angleRad);
	MMatrix2 rotate(0);
	rotate(0, 0) = cs;
	rotate(0, 1) = -sn;
	rotate(1, 0) = sn;
	rotate(0, 1) = cs;
	MMatrix2 rotateInv(0);
	rotateInv(0, 0) = cs;
	rotateInv(0, 1) = sn;
	rotateInv(1, 0) = -sn;
	rotateInv(0, 1) = cs;

	MMatrix2 covWRTB;
	covWRTB.isMult(velCov, rotate);
	covWRTB.preMult(rotateInv);

	return covWRTB;
}

double KI2DWithVelocity::getIVLength() {
	return ivLength;
}

double KI2DWithVelocity::getNumSDVel(double vx, double vy) {
	MVec2 diff;
	diff(0, 0) = vx - velVal(0, 0);
	diff(1, 0) = vy - velVal(1, 0);

	MRowVec2 diffTrans;
	diffTrans.isTranspose(diff);

	MMatrix2 velCovInv;
	velCovInv.isInverse(velCov);
	MRowVec2 val1;
	val1.isMult(diffTrans, velCovInv);
	MMatrix1 val2;
	val2.isMult(val1, diff);
	return val2(0, 0);
}

int KI2DWithVelocity::serialize(unsigned char *add, int base) {
	int ret = KalmanInfo2D::serialize(add, base);
	ret = velVal.serialize(add, ret);
	ret = velCov.serialize(add, ret);

	memcpy(add + ret, &numFrames, sizeof(int));
	ret += (int)sizeof(int);
	memcpy(add + ret, &lastBallMethod, sizeof(int));
	ret += (int)sizeof(int);

	float temp = (float) lastDist;
	memcpy(add + ret, &temp, sizeof(float));
	ret += (int)sizeof(float);
	temp = (float) lastHead;
	memcpy(add + ret, &temp, sizeof(float));
	ret += (int)sizeof(float);
	temp = (float) lastDistVar;
	memcpy(add + ret, &temp, sizeof(float));
	ret += (int)sizeof(float);
	temp = (float) lastHeadVar;
	memcpy(add + ret, &temp, sizeof(float));
	ret += (int)sizeof(float);

	ret = lastPos.serialize(add, ret);

	return ret;
}


int KI2DWithVelocity::unserialize(unsigned char *add, int base) {
	int ret = KalmanInfo2D::unserialize(add, base);
	ret = velVal.unserialize(add, ret);
	ret = velCov.unserialize(add, ret);

	memcpy(&numFrames, add + ret, sizeof(int));
	ret += (int)sizeof(int);
	memcpy(&lastBallMethod, add + ret, sizeof(int));
	ret += (int)sizeof(int);

	float temp;
	memcpy(&temp, add + ret, sizeof(float));
	lastDist = (double) temp;
	ret += (int)sizeof(float);
	memcpy(&temp, add + ret, sizeof(float));
	lastHead = (double) temp;
	ret += (int)sizeof(float);
	memcpy(&temp, add + ret, sizeof(float));
	lastDistVar = (double) temp;
	ret += (int)sizeof(float);
	memcpy(&temp, add + ret, sizeof(float));
	lastHeadVar = (double) temp;
	ret += (int)sizeof(float);

	ret = lastPos.unserialize(add, ret);

	velGlobalCur = false;
	velLocalCur = false;

	return ret;
}

////////////////////////////////////
// private KI2DWithVelocity methods
////////////////////////////////////

inline void KI2DWithVelocity::velNorm2Global() {
	gVelVal.setVector(vCART, velVal(0, 0), velVal(1, 0));
	velGlobalCur = true;
}

inline void KI2DWithVelocity::velNorm2Local() {
	lVelVal.setVector(vCART, velVal(0, 0), velVal(1, 0));
	lVelVal.rotate(90 - (gps->self()).h);
	velLocalCur = true;
}


