/*

   Copyright 2004 The University of New South Wales (UNSW) and National  
   ICT Australia (NICTA).

   This file is part of the 2004 team rUNSWift RoboCup entry.  You may  
   redistribute it and/or modify it under the terms of the GNU General  
   Public License as published by the Free Software Foundation; either  
   version 2 of the License, or (at your option) any later version as  
   modified below.  As the original licensors, we add the following  
   conditions to that license:

   In paragraph 2.b), the phrase "distribute or publish" should be  
   interpreted to include entry into a competition, and hence the source  
   of any derived work entered into a competition must be made available  
   to all parties involved in that competition under the terms of this  
   license.

   In addition, if the authors of a derived work publish any conference  
   proceedings, journal articles or other academic papers describing that  
   derived work, then appropriate academic citations to the original work  
   must be included in that publication.

   This rUNSWift source is distributed in the hope that it will be useful,  
   but WITHOUT ANY WARRANTY; without even the implied warranty of  
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
   General Public License for more details.

   You should have received a copy of the GNU General Public License along  
   with this source code; if not, write to the Free Software Foundation,  
   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 */

/*
 * Last modification background information
 * $Id: PyEmbed.cc 4362 2004-09-22 06:54:12Z kcph007 $
 *
 * Copyright (c) 2004 UNSW
 * All Rights Reserved.
 *
**/



#include "Python.h"
#include "PyEmbed.h"

#include <iostream>

#include "../share/IndicatorsDef.h"
#include "../share/Common.h"
#include "../vision/CommonSense.h"
#include "../share/FileUtility.h"

#ifdef COMPILE_ALL_CPP
#include "../behaviours/UNSW2004.h"
#endif // COMPILE_ALL_CPP

#include "../behaviours/Behaviours.h"
#include "../behaviours/pythonBehaviours.h"
#include "../behaviours/WalkingLearner.h"
#include "InfoManager.h"
#include "PointProjection.h"

#ifdef OFFLINE
#include <sys/time.h>
#endif // OFFLINE

using namespace std;

static PyObject *initVisionLinkModule(void);  // forward reference
static void addConstants(PyObject *module);

// see Python-Docs-2.3.3/ext/callingPython.html
// and Python-Docs-2.3.3/api/veryhigh.html

static char *VisionLinkModuleName = "VisionLink";
static char *BehaviourModuleName = "Behaviour";                                 

static const bool debugCouts = false;

static PyObject *behaviourModule = NULL;

static PyObject *ProcessFramePyFunc = NULL;
static PyObject *ProcessCommandPyFunc = NULL;

namespace PyBehaviours {
#ifndef OFFLINE
    Vision *transmission;
#endif	// OFFLINE
    VisualCortex *vision;
    GPS *gps;
    CommonSense *sensors;
    struct PWalkInfoStruct *PWalkInfo;

    void initializeBehaviours() {
        cout << "Initializing Python Behaviours" << endl;

        if (ProcessFramePyFunc != NULL) {
            Py_DECREF(ProcessFramePyFunc);
            ProcessFramePyFunc = NULL;
        }
        if (ProcessCommandPyFunc != NULL) {
            Py_DECREF(ProcessCommandPyFunc);
            ProcessCommandPyFunc = NULL;
        }
        if (Py_IsInitialized()) {
            cout << "Finalizing previous Python Behaviours" << endl;
            Py_Finalize();
        }

        Py_VerboseFlag += 2;

        Py_Initialize();

        cout << "Importing VisionLink" << endl;

        PyObject *module = initVisionLinkModule();

        addConstants(module);

        cout << "Importing " << BehaviourModuleName << endl;

        if ( (behaviourModule = PyImport_ImportModule(BehaviourModuleName) ) == NULL) {
            cout << "Error Importing Behaviours" << endl;
        }

        cout << "Done initializing Python Behaviours" << endl;
    }

    void processFrame() {
        if (ProcessFramePyFunc == NULL) {
            cout
                << "PyBehaviours::processFrame(): Python callback not initialized.  Unable to call it."
                << endl;
            return;
        }

        PyObject *args = Py_BuildValue("()");
        if (args == NULL) {
            cout << "PyBehaviours::processFrame(): Unable to build args" << endl;
            if (PyErr_Occurred() == NULL) {
                cout << "But no Python exception available!?!" << endl;
            }
            else {
                PyErr_Print();
            }
            return;
        }
        PyObject *result = PyEval_CallObject(ProcessFramePyFunc, args);
        Py_DECREF(args);
        if (result == NULL) {
            cout << "PyBehaviours::processFrame(): Got NULL from python call"
                << endl;
            if (PyErr_Occurred() == NULL) {
                cout << "But no Python exception available!?!" << endl;
            }
            else {
                PyErr_Print();
            }
        }
        else {
            Py_DECREF(result);
        }
    }

    void processCommand(const char *cmd) {
        if (ProcessCommandPyFunc == NULL) {
            cout
                << "PyBehaviours::processCommand(): Python callback not initialized.  Unable to call it."
                << endl;
            return;
        }

        PyObject *args = Py_BuildValue("(s)", cmd);
              
        if (args == NULL) {
            cout << "PyBehaviours::processCommand(): Unable to build args" << endl;
            if (PyErr_Occurred() == NULL) {
                cout << "But no Python exception available!?!" << endl;
            }
            else {
                PyErr_Print();
            }
            return;
        }
        PyObject *result = PyEval_CallObject(ProcessCommandPyFunc, args);
        Py_DECREF(args);
        if (result == NULL) {
            cout << "PyBehaviours::processCommand(): Got NULL from python call"
                << endl;
            if (PyErr_Occurred() == NULL) {
                cout << "But no Python exception available!?!" << endl;
            }
            else {
                PyErr_Print();
            }
        }
        else {
            Py_DECREF(result);
        }
    }

    void runString(char *cmd) {
        PyRun_SimpleString(cmd);
    }

    void deleteAllPYCFiles(){
        static const char *PyDir = "/MS/PyCode";
        vector<string> fileList = listAllFiles( PyDir );
        for (unsigned int i=0;i<fileList.size();i++){
            if (fileType(fileList[i]) == "PYC"){
                deleteFile((string(PyDir) + string("/") + fileList[i]).c_str());
            }
        }
    }

    void reloadModule(){
        if (behaviourModule == NULL){
            cout << "Error : PyBehaviour module is not initialized" << endl;
        }
        else{
            deleteAllPYCFiles();
            PyObject *oldModule = behaviourModule;
            behaviourModule = PyImport_ReloadModule(behaviourModule);
            if (behaviourModule == NULL){
                cout << "Error : New Python Behaviour module can not be loaded, revert to old module" << endl;
                behaviourModule = oldModule;
            }
            else {
                cout << "New python behaviours reloaded successfully" << endl;
            }
        }
    }

    //This code is acknowledged to Rick L. Ratzel
    //http://elmer.sourceforge.net/PyCon04/elmer_pycon04.html 
    //however, doesn't work, I ended up using sendMessageToBase, which is more flexible
    char *evaluateExpression(const char *expression){
        static const int BUFFER_SIZE = 500;
        char buffer[BUFFER_SIZE];

        PyObject* evalModule;
        PyObject* evalDict;
        PyObject* evalVal;
        char* retString;
        //sprintf(buffer, "result = 'foo' + ' bar'");
        sprintf(buffer, "result = `%s`", expression);
        cout << "Running " << buffer << endl;

        //PyRun_SimpleString( buffer);

        //evalModule = PyImport_AddModule( (const char*)"__main__" );
        evalDict = PyModule_GetDict( behaviourModule );
        /*
        evalVal = PyRun_String(buffer, 0, evalDict, evalDict);
        if (evalVal != NULL)
            cout << "Got result " << PyString_AsString(evalVal) << endl;
        else 
            cout << "Got NULL" << endl;
            */
        //evalDict = PyModule_GetDict( evalModule );
        evalVal = PyDict_GetItemString( evalDict, "result" );

        if( evalVal == NULL ) {
            return NULL;
        } else {
            /*
             * PyString_AsString returns char* repr of PyObject, which should
             * not be modified in any way...this should probably be copied for
             * safety
             */
            return PyString_AsString( evalVal );
        }
    }
}

/*
 * Start porting global constants into module
 */

static void addConstants(PyObject *module){
    PyObject *dic = PyModule_GetDict(module);

    PyObject *tmp;
    tmp = PyInt_FromLong(NUM_OF_EFFECTORS);
    PyDict_SetItemString(dic, "NUM_OF_EFFECTORS", tmp);
    Py_DECREF(tmp);

    tmp = PyInt_FromLong(NUM_SENSORS);
    PyDict_SetItemString(dic, "NUM_SENSORS", tmp);
    Py_DECREF(tmp);

    tmp = PyInt_FromLong(NUM_TEAM_MEMBER);
    PyDict_SetItemString(dic, "NUM_TEAM_MEMBER", tmp);
    Py_DECREF(tmp);

    tmp = PyFloat_FromDouble(BALL_RADIUS);
    PyDict_SetItemString(dic, "BALL_RADIUS", tmp);
    Py_DECREF(tmp);

    tmp = PyInt_FromLong(FIELD_WIDTH);
    PyDict_SetItemString(dic, "FIELD_WIDTH", tmp);
    Py_DECREF(tmp);

    tmp = PyInt_FromLong(FIELD_LENGTH);
    PyDict_SetItemString(dic, "FIELD_LENGTH", tmp);
    Py_DECREF(tmp);

    tmp = PyInt_FromLong(WALL_THICKNESS);
    PyDict_SetItemString(dic, "WALL_THICKNESS", tmp);
    Py_DECREF(tmp);

    tmp = PyInt_FromLong(GOAL_WIDTH);
    PyDict_SetItemString(dic, "GOAL_WIDTH", tmp);
    Py_DECREF(tmp);

    tmp = PyInt_FromLong(GOAL_LENGTH);
    PyDict_SetItemString(dic, "GOAL_LENGTH", tmp);
    Py_DECREF(tmp);
}

/*
 * Start porting static python functions
 */
                                               

using namespace PyBehaviours;

/*
 * Helper functions to convert types from C++ to Python
 */

static PyObject * PyObject_FromWMObject(const WMObj &obj){

    PyObject *t;

    t = PyTuple_New(5);
    PyTuple_SetItem(t, 0, PyFloat_FromDouble(obj.pos.x) );
    PyTuple_SetItem(t, 1, PyFloat_FromDouble(obj.pos.y) );
    PyTuple_SetItem(t, 2, PyFloat_FromDouble(obj.posVar) );
    PyTuple_SetItem(t, 3, PyFloat_FromDouble(obj.h));
    PyTuple_SetItem(t, 4, PyFloat_FromDouble(obj.hVar));
    
    return t;
}

static PyObject * PyObject_FromVector(const Vector &vector){

    PyObject *t;
                                               

    t = PyTuple_New(6);
    PyTuple_SetItem(t, 0, PyFloat_FromDouble(vector.x) );
    PyTuple_SetItem(t, 1, PyFloat_FromDouble(vector.y) );
    PyTuple_SetItem(t, 2, PyFloat_FromDouble(vector.d) );
    PyTuple_SetItem(t, 3, PyFloat_FromDouble(vector.theta));
    PyTuple_SetItem(t, 4, PyFloat_FromDouble(vector.head));
    PyTuple_SetItem(t, 5, PyFloat_FromDouble(vector.angleNorm));
    
    return t;
}

static PyObject * PyObject_FromCoordArray(const coordArray *coord){
    PyObject *t;

    t = PyTuple_New(8);
    for (int i=0;i<8;i++){
        PyObject *pos = PyTuple_New(2);
        PyTuple_SetItem(pos,0, PyInt_FromLong((*coord)[i][0]));
        PyTuple_SetItem(pos,1, PyInt_FromLong((*coord)[i][1]));
        PyTuple_SetItem(t,i,pos);
    }
    //TODO
    return t;
}

static PyObject * PyObject_From3x3Maxtrix(const MMatrix3 &maxtrix){
    PyObject *t;

    t = PyTuple_New(3);
    for (int i=0;i<3;i++){
        PyObject *row = PyTuple_New(3);

        for (int j=0;j<3;j++)
            PyTuple_SetItem(row, j, PyFloat_FromDouble(maxtrix(i,j)));
        PyTuple_SetItem(t,i,row);
    }
    //TODO
    return t;
}

static PyObject * VisionLink_dummy(PyObject * /*self*/,
        PyObject *args) {
    if (debugCouts) {
        cout << "oh no, your're runnig dummy function " << endl;
    }

    PyErr_Format(PyExc_RuntimeError,
            "%s not completely implemented in file %s at line %d.",
            __PRETTY_FUNCTION__,
            __FILE__,
            __LINE__);

    return NULL;
}

/* My static module */

static PyObject * VisionLink_setCallbacks(PyObject * /* self */,
        PyObject *args) {
    PyObject *result = NULL;
    PyObject *processFramePtr;
    PyObject *processCmdPtr;

    cout << "Setting Python VisionLink callbacks" << endl;

    if (PyArg_ParseTuple(args,
                "OO:setCallbacks",
                &processFramePtr,
                &processCmdPtr)) {
        if (!PyCallable_Check(processFramePtr) || !PyCallable_Check(processCmdPtr)) {
            PyErr_SetString(PyExc_TypeError, "parameter must be callable");
            return NULL;
        }
        Py_XINCREF(processFramePtr);         /* Add a reference to new callback */
        if (ProcessFramePyFunc != NULL) {
            Py_DECREF(ProcessFramePyFunc);  /* Dispose of previous callback */
        }
        ProcessFramePyFunc = processFramePtr;       /* Remember new callback */
        Py_XINCREF(processCmdPtr);         /* Add a reference to new callback */
        if (ProcessCommandPyFunc != NULL) {
            Py_DECREF(ProcessCommandPyFunc);  /* Dispose of previous callback */
        }
        ProcessCommandPyFunc = processCmdPtr;       /* Remember new callback */
        /* Boilerplate to return "None" */
        Py_INCREF(Py_None);
        result = Py_None;
    }
    return result;
}

/*  Set the atomic action. */
static PyObject * VisionLink_sendAtomicAction(PyObject * /*self*/,
        PyObject *args) {
    //struct AtomicAction thisAA;
    long cmd;

    if (debugCouts) {
        cout << "starting send of atomic action" << endl;
    }

    if (PyArg_ParseTuple(args,
                "lidddiddddddddiddd:sendAtomicAction",
                &cmd,
                &Behaviours::walkType ,
                &Behaviours::forward ,
                &Behaviours::left ,
                &Behaviours::turnCCW ,
                &Behaviours::PG ,
                &Behaviours::hF ,
                &Behaviours::hB,
                &Behaviours::hdF,
                &Behaviours::hdB,
                &Behaviours::ffO,
                &Behaviours::fsO,
                &Behaviours::bfO,
                &Behaviours::bsO,
                &Behaviours::headtype,
                &Behaviours::panx,
                &Behaviours::tilty,
                &Behaviours::cranez
                )) {

      // HACK: is the command is setting high / low gain, then continue
      // the motion, but hack the cmd back to aaParaWalk
        if (cmd == aaSetHighGain || cmd == aaSetLowGain) {
            AtomicAction aa ;
            aa.cmd      = cmd;
            //cout << "eWHICH JOINTS: "<< aa.walkType << endl;
#ifndef OFFLINE
            transmission->sendAA(aa);
#endif //OFFLINE
            Behaviours::acCmd = aaParaWalk;
        }
        else{
            Behaviours::acCmd = cmd;
        }
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}

/*  Set the LEDs, tail and ears. */
static PyObject * VisionLink_sendIndicators(PyObject * /*self*/, PyObject *args) {
    if (debugCouts) {
        cout << "starting sendIndicators" << endl;
    }

    int inds[NUM_OF_INDICATORS];

    if (PyTuple_Size(args) != NUM_OF_INDICATORS) {
        PyErr_Format(PyExc_RuntimeError,
                "Incorrect number of arguments to sendIndicators.  Should be %d",
                (int)
                NUM_OF_INDICATORS);                                               

        return NULL;
    }

    for (int i = 0; i < NUM_OF_INDICATORS; i++) {
        PyObject *thisInd = PyTuple_GetItem(args, i);
        if (thisInd == NULL)
            return NULL;
        inds[i] = PyInt_AsLong(thisInd);
        if (PyErr_Occurred()) {
            return NULL;
        }
    }	

#ifndef OFFLINE

    // Turn on the wireless indicator if a packet is received. ind[24] is the wireless indicator.
	if (transmission->gotTMWM) {
		inds[24] = IND_LED2_ON;
		transmission->gotTMWM = 0;
	} else {
		inds[24] = IND_LED2_OFF;
	}

	// If there are three or more teammates with valid world models turn the wireless indicator on.
	int tm = 0;
	for (int i=0 ; i < NUM_TEAM_MEMBER ; i++) {
		if (gps->tmObj(i).counter > 0) {
			tm++;
		}
	}
	if (tm >= 3) {
		inds[24] = IND_LED2_ON;
	} 
	
    transmission->sendIndicators(inds, sizeof(inds));

#endif // OFFLINE

    Py_INCREF(Py_None);
    return Py_None;
}

/* Send wireless back to base station, prefix by "****" */
static PyObject * VisionLink_sendMessageToBase(PyObject * /*self*/, PyObject *args) {

    char *message;
    if (PyArg_ParseTuple(args,
                "s:sendMessageToBase",&message)){
#ifndef OFFLINE
        transmission->sendWirelessMessage(message);
#endif // OFFLINE
    }
    else
        return NULL;

    Py_INCREF(Py_None);
    return Py_None;
}
/* Send my behaviours to teammates via gps */
static PyObject * VisionLink_sendMyBehaviourInfo(PyObject * /*self*/, PyObject *args) {

    double ballDist;
    bool isStriker;
    bool isGoalie;
    bool haveBall;
    bool kickedOff;
    bool isDefending;
    bool isBackingOff;
    interpBehav info ;
    if (PyArg_ParseTuple(args,
                "dbbbbbb:sendMyBehaviourInfo",&ballDist, &isStriker, &isGoalie, &haveBall, &kickedOff, &isDefending, &isBackingOff)){
        info.ballDist = (unsigned int) ballDist;
        info.amStriker = int(isStriker);
        info.amGoalie = int(isGoalie);
        info.haveBall = int(haveBall);
        info.kickedOff = int(kickedOff);
        info.defending = int(isDefending);
        info.backingOff = int(isBackingOff);
        gps->sharedBehaviour() = info.convert();
    }
    else
        return NULL;

    Py_INCREF(Py_None);
    return Py_None;
}

/*  Send our world model to our teammates. */
static PyObject * VisionLink_sendEnvironmentTeamMates(PyObject * /*self*/,
        PyObject *args) {
    if (debugCouts) {
        cout << "starting sendEnvironmentTeamMates" << endl;
    }

#ifndef OFFLINE
    transmission->shareEnvironment();
#endif // OFFLINE
    Py_INCREF(Py_None);
    return Py_None;
}

/*  Send the CPlane to the base station. */
static PyObject * VisionLink_sendCompressedCPlane(PyObject * /*self*/,
        PyObject *args) {
    if (debugCouts) {
        cout << "starting sendCompressedCPlane" << endl;
    }

#ifndef OFFLINE
    transmission->sendCompressedCPlane();
#endif // OFFLINE
    Py_INCREF(Py_None);
    return Py_None;
}

/*  Send the OPlane to the base station. */
static PyObject * VisionLink_sendOPlane(PyObject * /*self*/, PyObject *args) {
    if (debugCouts) {
        cout << "starting sendOPlane" << endl;
    }

#ifndef OFFLINE
    transmission->sendOPlane();
#endif // OFFLINE
    Py_INCREF(Py_None);
    return Py_None;
}

/*  Send the world model to the base station. */
static PyObject * VisionLink_sendEnvironmentBaseStation(PyObject * /*self*/,
        PyObject *args) {
    if (debugCouts) {
        cout << "starting sendEnvironmentBaseStation" << endl;
    }

#ifndef OFFLINE
    transmission->displayEnvironment();
#endif // OFFLINE
    Py_INCREF(Py_None);
    return Py_None;
}

/*  Send the YUV plane to the base station. */
static PyObject * VisionLink_sendYUVPlane(PyObject * /*self*/, PyObject *args) {
    if (debugCouts) {
        cout << "starting sendYUVPlane" << endl;
    }

#ifndef OFFLINE
    transmission->sendYUVPlane();
#endif // OFFLINE
    Py_INCREF(Py_None);
    return Py_None;
}

/*  Get the robot's battery level. */
static PyObject * VisionLink_getBatteryLevel(PyObject * /*self*/,
        PyObject *args) {
    if (debugCouts) {
        cout << "starting getBatteryLevel" << endl;
    }

    int batLevel = 50;
#ifndef OFFLINE
    OStatus result;
    OPowerStatus currentStatus;

    result = OPENR::GetPowerStatus(&currentStatus);
    if (result == oSUCCESS) {
        batLevel = currentStatus.remainingCapacity;
        if (currentStatus.robotStatus & orsbBATTERY_CAPACITY_LOW) {
            batLevel *= -1;
        }
    }
    else {
        PyErr_Format(PyExc_RuntimeError,
                "Error getting power status: %d",
                (int) result);
        return NULL;
    }
#endif // OFFLINE
    return PyInt_FromLong(batLevel);
}

static PyObject * VisionLink_getBatteryCurrent(PyObject * /*self*/,
        PyObject *args) {

#ifndef OFFLINE
    OStatus result;
    OPowerStatus currentStatus;

    result = OPENR::GetPowerStatus(&currentStatus);
    if (result != oSUCCESS) {
        PyErr_Format(PyExc_RuntimeError,
                "Error getting power status: %d",
                (int) result);
        return NULL;
    }
    return PyInt_FromLong(-currentStatus.current);
#else // OFFLINE
    return PyInt_FromLong(1500);
#endif // OFFLINE
}

/*  Get the state of the robot's touch sensors. */
static PyObject * VisionLink_getPressSensorCount(PyObject * /*self*/,
        PyObject *arg) {
    if (debugCouts) {
        cout << "starting getPressSensorCount" << endl;
    }

    int sensorID = PyInt_AsLong(arg);

    if (PyErr_Occurred()) {
        return NULL;
    }

    if (sensorID <0 || sensorID> NUM_PRESSES) {
        PyErr_Format(PyExc_RuntimeError, "Bad press sensor ID: %d", sensorID);
        return NULL;
    }

    return PyInt_FromLong(sensors->press_sensor[sensorID].count);
}

/*  Have we fallen over? */
static PyObject * VisionLink_getFallenState(PyObject * /*self*/, PyObject *args) {
    if (debugCouts) {
        cout << "starting getFallenState" << endl;
    }

    return PyInt_FromLong(sensors->hasFallen);
}

/*  Tell GPS which goal is ours. */
static PyObject * VisionLink_setGPSGoal(PyObject * /*self*/, PyObject *arg) {
    if (debugCouts) {
        cout << "starting setGPSGoal" << endl;
    }

    bool goalID = (bool)PyInt_AsLong(arg);

    if (PyErr_Occurred()) {
        return NULL;
    }

    gps->SetGoals(goalID);
    // might want to fix this in future...
#ifndef OFFLINE
    transmission->sendPlayerNumber(Behaviours::PLAYER_NUMBER, goalID);
#endif // OFFLINE

    Py_INCREF(Py_None);
    return Py_None;
}

/*  Pause GPS updates.  Useful when penalised. */                                               

static PyObject * VisionLink_setGPSPaused(PyObject * /*self*/, PyObject *arg) {
    if (debugCouts) {
        cout << "starting setGPSPaused" << endl;
    }

    int penal = PyInt_AsLong(arg);

    if (PyErr_Occurred()) {
        return NULL;
    }

    gps->setPenalized((penal != 0));

    Py_INCREF(Py_None);
    return Py_None;
}


/*  Get information on the ball. */
static PyObject * VisionLink_getGPSBallInfo(PyObject * /*self*/, PyObject *args) {
    int context;

    if (!PyArg_ParseTuple(args, "i", &context) || (context != GLOBAL && context != LOCAL) ) {
        Py_XINCREF(Py_None);
        return Py_None;
    }

//Temporary
    return PyObject_FromVector(gps->getBall(context));

//Eventually
//    return newPVector(&gps->getBall(context));
    
}

// Ted: ???????? We won't use ball velocity in 2004 ......
/*  Get information on the ball's velocity. */
static PyObject * VisionLink_getGPSBallVInfo(PyObject * /*self*/,
        PyObject *args) {
    int context;
    if (!PyArg_ParseTuple(args, "i", &context) || (context != GLOBAL && context != LOCAL) ) {
        Py_XINCREF(Py_None);
        return Py_None;
    }

    return PyObject_FromVector(gps->getVBall(context));
}

/*  Get information on an opponent. */
static PyObject * VisionLink_getGPSOpponentInfo(PyObject * /*self*/,
        PyObject *args) {
    int context, opponentNumber;
    if (!PyArg_ParseTuple(args, "ii", &opponentNumber, &context) || (context != GLOBAL && context != LOCAL) || opponentNumber<0 || opponentNumber > 3 ) {
        Py_XINCREF(Py_None);
        return Py_None;
    }
    return PyObject_FromVector(gps->getOppPos(opponentNumber, context)) ;
}

static PyObject * VisionLink_getGPSOpponentCoxMax(PyObject * /*self*/,
        PyObject *args) {
    int opponentNumber;
    if (!PyArg_ParseTuple(args, "i", &opponentNumber) || opponentNumber < 0 || opponentNumber > 3 ) {
        Py_XINCREF(Py_None);
        return Py_None;
    }
    return PyFloat_FromDouble(gps->getOppCovMax(opponentNumber)) ;
}

static PyObject * VisionLink_getGPSBallMaxVar(PyObject * /*self*/,
        PyObject *args) {
    return PyFloat_FromDouble(gps->getBallMaxVar()) ;
}

static PyObject * VisionLink_getGPSVBallMaxVar(PyObject * /*self*/,
        PyObject *args) {
    return PyFloat_FromDouble(gps->getVBallMaxVar()) ;
}

/*  Get information about a teammates. */
static PyObject * VisionLink_getGPSTeammateInfo(PyObject * /*self*/,
        PyObject *args) {
    int context, teammateNumber;
    if (!PyArg_ParseTuple(args, "ii", &context, &teammateNumber) || (context != GLOBAL && context != LOCAL) || teammateNumber< 1 || teammateNumber > 4 ) {
	cout << "VisionLink_getGPSTeammateInfo from C++, wrong input argument!" << endl;
        Py_XINCREF(Py_None);
        return Py_None;
    }
    switch (context){
        case GLOBAL:
            return PyObject_FromWMObject(gps->teammate(teammateNumber, 'g')) ;
        case LOCAL:
            return PyObject_FromWMObject(gps->teammate(teammateNumber, 'l')) ;
        default:
            return NULL;
    }
}

/*  Get information about the opponent's goal. */
static PyObject * VisionLink_getOppGoalInfo(PyObject * /*self*/, PyObject *args) {
    return PyObject_FromWMObject(gps->tGoal());
}

/*  Get information about our own goal. */
static PyObject * VisionLink_getOwnGoalInfo(PyObject * /*self*/, PyObject *args) {
    return PyObject_FromWMObject(gps->oGoal());
}

/*  Get coordinate array (see GPS::coord), depending on which team it is in . */
static PyObject * VisionLink_getGPSCoordArray(PyObject * /*self*/, PyObject * /*args */) {
    return PyObject_FromCoordArray(gps->coord);
}


static PyObject * PyObject_FromVisualObject(VisualObject &obj){

    PyObject *t;
    const WMObj& r = gps->self();
    double d = obj.getFlatDistance();
    double globalx,globaly;

    calPosition (r.pos.x, r.pos.y, r.h, d, obj.h , &globalx, &globaly );

    t = PyTuple_New(14);
    PyTuple_SetItem(t, 0, PyInt_FromLong(obj.cf));
    PyTuple_SetItem(t, 1, PyFloat_FromDouble(obj.head));
    PyTuple_SetItem(t, 2, PyFloat_FromDouble(d)); //this distance is weird when fireball
    PyTuple_SetItem(t, 3, PyFloat_FromDouble(obj.var));
    PyTuple_SetItem(t, 4, PyFloat_FromDouble(globalx));
    PyTuple_SetItem(t, 5, PyFloat_FromDouble(globaly));

    PyTuple_SetItem(t, 6, PyFloat_FromDouble(obj.cx) );
    PyTuple_SetItem(t, 7, PyFloat_FromDouble(obj.cy) );
    PyTuple_SetItem(t, 8, PyFloat_FromDouble(obj.height));
    PyTuple_SetItem(t, 9, PyFloat_FromDouble(obj.width));
    PyTuple_SetItem(t, 10, PyFloat_FromDouble(obj.imgHead));
    PyTuple_SetItem(t, 11, PyFloat_FromDouble(obj.imgElev));
    PyTuple_SetItem(t, 12, PyFloat_FromDouble(obj.d)); //this distance is better,esp fireball
    PyTuple_SetItem(t, 13, PyFloat_FromDouble(obj.elev)); //this distance is better,esp fireball

    
    return t;
}

static PyObject * VisionLink_getVisualObject(PyObject * /*self*/,
        PyObject *args) {
    int vobIndex;
    if (PyArg_ParseTuple(args,
                "i:getVisualObject",
                &vobIndex) && vobIndex >= 0 && vobIndex < VOB_COUNT  ) {
        return PyObject_FromVisualObject(vision->vob[vobIndex]);
    }
    else {
        Py_XINCREF(Py_None);
        return Py_None;
    }
}


/*  Get information about the ball sent by our teammates. */
static PyObject * VisionLink_getSharedBallInfo(PyObject * /*self*/,
        PyObject *args) {
    PyObject *t = PyTuple_New(4);
    PyTuple_SetItem(t, 0, PyFloat_FromDouble( gps->shareBallx));
    PyTuple_SetItem(t, 1, PyFloat_FromDouble( gps->shareBally));
    PyTuple_SetItem(t, 2, PyFloat_FromDouble( gps->shareBallvar));
    PyTuple_SetItem(t, 3, PyInt_FromLong( gps->sbRobotNum));

    return t;
}

static PyObject * VisionLink_getJointSensor(PyObject * /*self*/,
        PyObject *args) {
    if (debugCouts) {
        cout << "starting getJointSensor" << endl;
    }

    int jointEnum;
    if (PyArg_ParseTuple(args,
                "i:getJointSensor",
                &jointEnum)) {
        return PyFloat_FromDouble(gInfoManager.getJointSensor((JointEnum)jointEnum) );
    }
    else {
        Py_XINCREF(Py_None);
        return Py_None;
    }
}

static PyObject * VisionLink_getAnySensor(PyObject * /*self*/,
        PyObject *args) {
    if (debugCouts) {
        cout << "starting getJointSensor" << endl;
    }
    int sensorIndex;
    if (PyArg_ParseTuple(args,
                "i:getAnySensor",
                &sensorIndex)) {
        return PyLong_FromLong(gInfoManager.getAnySensor(sensorIndex) );
    }
    else {
        Py_XINCREF(Py_None);
        return Py_None;
    }
}

static PyObject * VisionLink_getJointPWMDuty(PyObject * /*self*/,
        PyObject *args) {
    if (debugCouts) {
        cout << "starting getJointPWMDuty" << endl;
    }

    int jointEnum;
    if (PyArg_ParseTuple(args,
                "i:getJointPWMDuty",
                &jointEnum)) {
        return PyLong_FromLong(gInfoManager.getJointPWMDuty((JointEnum)jointEnum) );
    }
    else {
        Py_XINCREF(Py_None);
        return Py_None;
    }
}

static PyObject * VisionLink_getPWalkInfo(PyObject * /*self*/,
        PyObject *args) {

    PyObject *ret;

    ret = PyTuple_New(10);
    PyTuple_SetItem(ret, 0, PyInt_FromLong(*PWalkInfo->currentWalkType));
    PyTuple_SetItem(ret, 1, PyInt_FromLong(*PWalkInfo->currentStep));
    PyTuple_SetItem(ret, 2, PyInt_FromLong(*PWalkInfo->currentPG));
    PyTuple_SetItem(ret, 3, PyInt_FromLong(*PWalkInfo->leading_leg));
    PyTuple_SetItem(ret, 4, PyFloat_FromDouble(*PWalkInfo->frontShoulderHeight));
    PyTuple_SetItem(ret, 5, PyFloat_FromDouble(*PWalkInfo->bodyTilt));
    PyTuple_SetItem(ret, 6, PyFloat_FromDouble(*PWalkInfo->desiredPan));
    PyTuple_SetItem(ret, 7, PyFloat_FromDouble(*PWalkInfo->desiredTilt));
    PyTuple_SetItem(ret, 8, PyFloat_FromDouble(*PWalkInfo->desiredCrane));
    PyTuple_SetItem(ret, 9, PyInt_FromLong(*PWalkInfo->isHighGain));

    return ret;
}

static PyObject * VisionLink_getSelfLocation(PyObject * /*self*/,
        PyObject * /*args*/ ) {
    if (debugCouts) {
        cout << "starting getSelfLocation" << endl;
    }

    PyObject *t;

    t = PyTuple_New(4);
    PyTuple_SetItem(t, 0, PyFloat_FromDouble(0.0));
    PyTuple_SetItem(t, 1, PyFloat_FromDouble(gInfoManager.getSelfLocation().pos.x));
    PyTuple_SetItem(t, 2, PyFloat_FromDouble(gInfoManager.getSelfLocation().pos.y));
    PyTuple_SetItem(t, 3, PyFloat_FromDouble(gInfoManager.getSelfLocation().h));

    return t;
}

/*  Get our location info. */
static PyObject * VisionLink_getGPSSelfInfo(PyObject * /*self*/, PyObject *args) {
    return PyObject_FromWMObject(gps->self());
}

/*  Get self covariane matrix (3x3). */
static PyObject * VisionLink_getGPSSelfCovariance(PyObject * /*self*/, PyObject *args) {
    return PyObject_From3x3Maxtrix(gps->selfCov());
}

static PyObject * VisionLink_getBallDistance(PyObject * /*self*/,
        PyObject * /*args*/ ) {
    return PyFloat_FromDouble(gInfoManager.getBallDistance());
}

static PyObject * VisionLink_getBallHeading(PyObject * /*self*/,
        PyObject * /*args*/ ) {
    return PyFloat_FromDouble(gInfoManager.getBallHeading());
}

static PyObject * VisionLink_getBallConfidence(PyObject * /*self*/,
        PyObject * /*args*/ ) {
    return PyFloat_FromDouble(gInfoManager.getBallConfidence());
}

static PyObject * VisionLink_getBallLocation(PyObject * /*self*/,
        PyObject * /*args*/ ) {
    if (debugCouts) {
        cout << "starting getBallLocationGlobal" << endl;
    }

    PyObject *t;

    t = PyTuple_New(4);
    PyTuple_SetItem(t, 0, PyFloat_FromDouble(gInfoManager.getBallConfidence()));
    PyTuple_SetItem(t, 1, PyFloat_FromDouble(gInfoManager.getBallLocation().x));
    PyTuple_SetItem(t, 2, PyFloat_FromDouble(gInfoManager.getBallLocation().y));
    PyTuple_SetItem(t, 3, PyFloat_FromDouble(gInfoManager.getBallLocation().d));

    return t;
}

static PyObject * VisionLink_getCurrentPreviousFrameID(PyObject * /*self*/, PyObject * /*args*/ ) {
    PyObject *t;
    t = PyTuple_New(2);
    PyTuple_SetItem(t, 0, PyInt_FromLong(Behaviours::currentFrameID));
    PyTuple_SetItem(t, 1, PyInt_FromLong(Behaviours::lastFrameID));
    return t;
}

static PyObject * VisionLink_getCurrentTime(PyObject * /*self*/, PyObject * /*args*/ ) {
    PyObject *t;
    t = PyTuple_New(2);
#ifndef OFFLINE
    SystemTime curTime;
    GetSystemTime(&curTime);
    PyTuple_SetItem(t, 0, PyInt_FromLong(curTime.seconds));
    PyTuple_SetItem(t, 1, PyInt_FromLong(curTime.useconds));
#else
	struct timeval tval;
	gettimeofday(&tval, NULL);
    PyTuple_SetItem(t, 0, PyInt_FromLong(tval.tv_sec));
    PyTuple_SetItem(t, 1, PyInt_FromLong(tval.tv_usec));
#endif
    return t;
}

static PyObject * VisionLink_setUseBeacons(PyObject * /*self*/, PyObject *args) {
	bool update;
	PyArg_ParseTuple(args, "i:setIfUseBeacons", &update);
    vision->lookForBeaconsAndGoals = update;
	cout << "Setting use beacons to: " << vision->lookForBeaconsAndGoals << endl;	
    Py_XINCREF(Py_None);
    return Py_None;
}

static const int FIRE_PAW_KICK_AUTO_BASIC = 0;
static const int GET_BEHIND_BALL          = 1;
static const int RUN_BEHIND_BALL          = 2;
static const int LOCATE_BALL              = 3;

static PyObject * VisionLink_getWirelessTeammateInfo(PyObject * /*self*/, PyObject *args) {
	PyObject *p;
	p = PyTuple_New(32);
             
	// Valid/invalid teammates.
	PyTuple_SetItem(p, 0, PyInt_FromLong(gps->tmObj(0).counter));               
	PyTuple_SetItem(p, 1, PyInt_FromLong(gps->tmObj(1).counter));
    PyTuple_SetItem(p, 2, PyInt_FromLong(gps->tmObj(2).counter));
    PyTuple_SetItem(p, 3, PyInt_FromLong(gps->tmObj(3).counter));
                            
    // Defending teammate?
    PyTuple_SetItem(p, 4, PyInt_FromLong((((interpBehav) gps->tmObj(0).behavioursVal)).defending));
    PyTuple_SetItem(p, 5, PyInt_FromLong((((interpBehav) gps->tmObj(1).behavioursVal)).defending));
    PyTuple_SetItem(p, 6, PyInt_FromLong((((interpBehav) gps->tmObj(2).behavioursVal)).defending));
    PyTuple_SetItem(p, 7, PyInt_FromLong((((interpBehav) gps->tmObj(3).behavioursVal)).defending));

    // Striker teammate?
    PyTuple_SetItem(p, 8,  PyInt_FromLong((((interpBehav) gps->tmObj(0).behavioursVal)).amStriker));
    PyTuple_SetItem(p, 9,  PyInt_FromLong((((interpBehav) gps->tmObj(1).behavioursVal)).amStriker));
    PyTuple_SetItem(p, 10, PyInt_FromLong((((interpBehav) gps->tmObj(2).behavioursVal)).amStriker));
    PyTuple_SetItem(p, 11, PyInt_FromLong((((interpBehav) gps->tmObj(3).behavioursVal)).amStriker));

    // Teammate ball distance.
    PyTuple_SetItem(p, 12, PyInt_FromLong((((interpBehav) gps->tmObj(0).behavioursVal)).ballDist));
    PyTuple_SetItem(p, 13, PyInt_FromLong((((interpBehav) gps->tmObj(1).behavioursVal)).ballDist));
    PyTuple_SetItem(p, 14, PyInt_FromLong((((interpBehav) gps->tmObj(2).behavioursVal)).ballDist));
    PyTuple_SetItem(p, 15, PyInt_FromLong((((interpBehav) gps->tmObj(3).behavioursVal)).ballDist));

	// Goalie or forward?
	PyTuple_SetItem(p, 16, PyInt_FromLong((((interpBehav) gps->tmObj(0).behavioursVal)).amGoalie));
	PyTuple_SetItem(p, 17, PyInt_FromLong((((interpBehav) gps->tmObj(1).behavioursVal)).amGoalie));
	PyTuple_SetItem(p, 18, PyInt_FromLong((((interpBehav) gps->tmObj(2).behavioursVal)).amGoalie));
	PyTuple_SetItem(p, 19, PyInt_FromLong((((interpBehav) gps->tmObj(3).behavioursVal)).amGoalie));
                
	// Teammate have ball?
	PyTuple_SetItem(p, 20, PyInt_FromLong((((interpBehav) gps->tmObj(0).behavioursVal)).haveBall));
	PyTuple_SetItem(p, 21, PyInt_FromLong((((interpBehav) gps->tmObj(1).behavioursVal)).haveBall));
	PyTuple_SetItem(p, 22, PyInt_FromLong((((interpBehav) gps->tmObj(2).behavioursVal)).haveBall));
	PyTuple_SetItem(p, 23, PyInt_FromLong((((interpBehav) gps->tmObj(3).behavioursVal)).haveBall));

	// Teammate backingOff?
	PyTuple_SetItem(p, 24, PyInt_FromLong((((interpBehav) gps->tmObj(0).behavioursVal)).backingOff));
	PyTuple_SetItem(p, 25, PyInt_FromLong((((interpBehav) gps->tmObj(1).behavioursVal)).backingOff));
	PyTuple_SetItem(p, 26, PyInt_FromLong((((interpBehav) gps->tmObj(2).behavioursVal)).backingOff));
	PyTuple_SetItem(p, 27, PyInt_FromLong((((interpBehav) gps->tmObj(3).behavioursVal)).backingOff));

	// Teammate kickOff?
	PyTuple_SetItem(p, 28, PyInt_FromLong((((interpBehav) gps->tmObj(0).behavioursVal)).kickedOff));
	PyTuple_SetItem(p, 29, PyInt_FromLong((((interpBehav) gps->tmObj(1).behavioursVal)).kickedOff));
	PyTuple_SetItem(p, 30, PyInt_FromLong((((interpBehav) gps->tmObj(2).behavioursVal)).kickedOff));
	PyTuple_SetItem(p, 31, PyInt_FromLong((((interpBehav) gps->tmObj(3).behavioursVal)).kickedOff));
                                
	return p;
}

static PyObject * VisionLink_doBasicBehaviour(PyObject * /*self*/, PyObject *args) {
    int action;
    if (PyArg_ParseTuple(args, "i:doBasicBehaviour", &action)) {
        switch (action){
            case FIRE_PAW_KICK_AUTO_BASIC:
                Behaviours::firePawKick(Behaviours::FIRE_PAWKICK_AUTO);
                break;
            case GET_BEHIND_BALL:
                PythonBehaviours::getBehindBall();
                break;
            case RUN_BEHIND_BALL:
                PythonBehaviours::runBehindBall();
                break;
            case LOCATE_BALL:
                PythonBehaviours::locateBall();
                break;
            default:
                cout << "Unknown basic behaviour" << endl;
        }
    }
    Py_XINCREF(Py_None);
    return Py_None;
}

/* Get the project for ball. When this function is called, 
   it is assumed that the robot can see the ball. */
static PyObject * VisionLink_getProjectedBall(PyObject * /*self*/, PyObject *args) {    
    // Track visual ball.
    int inpoints[2] = {((int) (vision->vob[vobBall].cx) - WIDTH / 2), -1 * ((int) (vision->vob[vobBall].misc) - HEIGHT / 2)};
    double outpoints[2] = {-1, -1};
    vision->projectPoints(inpoints, 1, outpoints, 0);
    double ballx = -outpoints[0];
    double bally = outpoints[1];

    PyObject *t;
    t = PyTuple_New(2);
    PyTuple_SetItem(t, 0, PyFloat_FromDouble(ballx));
    PyTuple_SetItem(t, 1, PyFloat_FromDouble(bally));
    return t;
}

/* Project a point from the vision frame onto the ground. */
static PyObject * VisionLink_getProjectedPoint(PyObject * /*self*/, PyObject *args) {    

	int inpoints[2];

    if (PyArg_ParseTuple(args,
                "ii:getProjectedPoint",
                &(inpoints[0]), &(inpoints[1]))) {

		double outpoints[2] = {-1, -1};
		vision->projectPoints(inpoints, 1, outpoints, 0);
	
		PyObject *t;
		t = PyTuple_New(2);
		PyTuple_SetItem(t, 0, PyFloat_FromDouble(-outpoints[0]));
		PyTuple_SetItem(t, 1, PyFloat_FromDouble(outpoints[1]));
		return t;
    } else 
        return NULL;

}

static PyObject * VisionLink_PointToElevation(PyObject * /*self*/, PyObject *args) {    

    double y;
    if (PyArg_ParseTuple(args,
                "d:PointToElevation",
                &y)) {
        //cout << "Got " << y << "  " << PointToElevation(y) << endl;
		return PyFloat_FromDouble(PointToElevation(y));
    } else 
        return NULL;

}

static PyObject * VisionLink_setCPlaneMode(PyObject * /*self*/, PyObject *args) {    

    int mode;
    if (PyArg_ParseTuple(args,
                "i:setCPlaneMode",
                &mode)) {
        cout << "Python Set CPlaneMode : " << mode << endl;
#ifndef OFFLINE
        sendCPlaneMode = mode;
#endif
        Py_XINCREF(Py_None);
        return Py_None;
    } else 
        return NULL;

}

static PyObject * VisionLink_PointToHeading(PyObject * /*self*/, PyObject *args) {    

    double x;
    if (PyArg_ParseTuple(args,
                "d:PointToHeading",
                &x)) {

        //cout << "Got " << x << "  " << PointToHeading(x) << endl;
		return PyFloat_FromDouble(PointToHeading(x));
    } else 
        return NULL;

}

static PyObject * VisionLink_getRedBlobInfo(PyObject * /*self*/, PyObject *args) {    

    PyObject *t;
    t = PyTuple_New(2);
    int x,y;
    gInfoManager.getAverageBlobInfo(x,y);
    PyTuple_SetItem(t, 0, PyInt_FromLong(x));
    PyTuple_SetItem(t, 1, PyInt_FromLong(y));
    return t;
}

static PyObject * VisionLink_getTestingInfo(PyObject * /*self*/, PyObject *args) {    
    /*
    //get total area of red
    int sumArea = 0;
    for (int i=0;i< vision->blobs->count(cROBOT_RED) ; i++){
        BlobInfo& redBlob = vision->blobs->getBlob(cROBOT_RED,i);
        sumArea += redBlob.area;
    }
    return PyInt_FromLong( sumArea );
    */
    
    int height = 0;
    //get dog's vision blob height
    for (int i=vobRedDog;i<=vobRedDog4;i++){
        if (vision->vob[i].height > 1){
            cout << "Object " << i << "[" << i - vobRedDog + 1 << "]" << vision->vob[i].height << endl;
            height = int(vision->vob[i].height);
        }
    }

    return PyInt_FromLong( height );
}

static PyObject * VisionLink_setWalkLearningParameters(PyObject * /*self*/, PyObject *args) {
    const char *paraStr;
    if (PyArg_ParseTuple(args,"s:setWalkLearningParameters",&paraStr)){
        WalkingLearner::setWalkLearningParameters(paraStr);
    }
    Py_XINCREF(Py_None);
    return Py_None;
}

static PyObject * VisionLink_getTeamColor (PyObject * /*self*/,
        PyObject * /*args*/ ) {
  if ( gps->targetGoal == vobYellowGoal ) {
    return PyInt_FromLong(0);
  }
  else {
    return PyInt_FromLong(1);
  }
}

static PyObject * VisionLink_getMyPlayerNum (PyObject * /*self*/,
        PyObject * /*args*/ ) {
    return PyInt_FromLong(Behaviours::PLAYER_NUMBER);
}

static PyObject * VisionLink_getGPSTeammateBallInfo (PyObject * /*self*/,
        PyObject *args) {
    int tmIndex;
    if (PyArg_ParseTuple(args,
                "i:getGPSTeammateBallInfo",
                &tmIndex) && tmIndex >= 0 && tmIndex < NUM_TEAM_MEMBER ) {
        return PyObject_FromWMObject(gps->tmBallObj(tmIndex));
    }
    else {
        Py_XINCREF(Py_None);
        return Py_None;
    }
}

static PyObject * VisionLink_printCPlusPlus(PyObject * /*self*/, PyObject *args) {
    int selection;
#ifdef COMPILE_ALL_CPP
    if (PyArg_ParseTuple(args, "i:printCPlusPlus", &selection)) {
        cout << "Selection: " << selection << endl;
        switch (selection){
            // Print results from vision strikerpoint.
            case 0:
                if (vision->vob[0].cf == 0)
                    cout << "Vision_Link_printCPlusPlus ball confidence problem!" << endl;
                else {
                    UNSW2004::determineBallSource();
                    double ptx, pty;
                    UNSW2004::getUNSW2004StrikerPoint(&ptx, &pty);
                    cout << "C++ striker point x: "  << ptx << endl;
                    cout << "C++ strriker point y: " << pty << endl;
                    UNSW2004::doUNSW2004StrikerForward(ptx, pty);
                    cout << "C++ striker forward: " << Behaviours::forward << endl;     
                    cout << "C++ striker turnCCW: " << Behaviours::turnCCW << endl;
                    cout << "C++ striker left:    " << Behaviours::left    << endl;
                    Behaviours::forward = 0;
                    Behaviours::turnCCW = 0;
                    Behaviours::left    = 0;
                }               
                break;          
            // Print GPS ball values for comparision.
            case 1:
                cout << "gps->getBall(GLOBAL).x: " << gps->getBall(GLOBAL).x << endl;
                cout << "gps->getBall(GLOBAL).y: " << gps->getBall(GLOBAL).y << endl;
                cout << "gps->getBall(GLOBAL).head: " << gps->getBall(GLOBAL).head << endl;
                break;
			case 2:
				Behaviours::saGetBehindBall(40, 90, 1);
				cout << "C++ behind forward: " << Behaviours::forward << endl;
				cout << "C++ behind turnCCW: " << Behaviours::turnCCW << endl;
				cout << "C++ behind left:    " << Behaviours::left    << endl;
				//Behaviours::forward = 0;
				//Behaviours::turnCCW = 0;
				//Behaviours::left    = 0;
				break;
			case 3:
				UNSW2004::determineBallSource();
				UNSW2004::doUNSW2004TrackVisualBall();
				UNSW2004::doUNSW2004Defend();
				cout << "C++ forward: " << Behaviours::forward << endl;
				cout << "C++ turnCCW: " << Behaviours::turnCCW << endl;
				cout << "C++ left:    " << Behaviours::left    << endl;
                Behaviours::forward = 0;
                Behaviours::turnCCW = 0;
                Behaviours::left    = 0;
				break;					
            default:
                cout << "PrintCPlusCPlus() got an invalid selection: " << selection << "!" << endl; 
        }
    }
#endif //COMPILE_ALL_CPP
    Py_XINCREF(Py_None);
    return Py_None;
}

// This wraps the voak getHeadingToBestGap(&min,&max), which determines
// the best range for goal scoring
static PyObject * VisionLink_getHeadingToBestGap (PyObject * /*self*/,
        PyObject *args) {

    PyObject *ret;
    double min, max;

    Behaviours::getHeadingToBestGap (&min,&max);

    ret = PyTuple_New(2);
    PyTuple_SetItem(ret, 0, PyFloat_FromDouble(min));
    PyTuple_SetItem(ret, 1, PyFloat_FromDouble(max));

    return ret;
}

static PyObject * VisionLink_getIfGPSMotionUpdate(PyObject * /*self*/, PyObject *args) {
	return PyInt_FromLong(gps->doGPSMotionUpdate);
}

static PyObject * VisionLink_getIfGPSVisionUpdate(PyObject * /*self*/, PyObject *args) {
	return PyInt_FromLong(gps->doGPSVisionUpdate);
}

static PyObject * VisionLink_setIfGPSMotionUpdate(PyObject * /*self*/, PyObject *args) {
	bool update;
	PyArg_ParseTuple(args, "i:setIfGpsMotionUpdate", &update);
    gps->doGPSMotionUpdate = update;
	cout << "Setting gps motion update to: " << gps->doGPSMotionUpdate << endl;	
    Py_XINCREF(Py_None);
    return Py_None;
}

static PyObject * VisionLink_setIfGPSVisionUpdate(PyObject * /*self*/, PyObject *args) {
	bool update;
	PyArg_ParseTuple(args, "i:setIfGpsVisionUpdate", &update);
    gps->doGPSVisionUpdate = update;
	cout << "Setting gps vision update to: " << gps->doGPSVisionUpdate << endl;	
    Py_XINCREF(Py_None);
    return Py_None;
}

static PyObject * VisionLink_setIfGPSOpponentUpdate(PyObject * /*self*/, PyObject *args) {
	bool update;
	PyArg_ParseTuple(args, "i:setIfGpsOpponentUpdate", &update);
    gps->doGPSOpponentUpdate = update;
    Py_XINCREF(Py_None);
    return Py_None;
}

static PyObject * VisionLink_setGlobalMaxEdge(PyObject * /*self*/, PyObject *args) {
    gps->GPSGlobalMaxEdgeSet(vision);
    return Py_None;
}

static PyObject * VisionLink_resetEdgeMap(PyObject * /*self*/, PyObject *args) {
    vision->invalidateEdgeMap();
    return Py_None;
}

static PyObject * VisionLink_setAllowPinkMapping(PyObject * /*self*/, PyObject *args) {
	bool update;
	PyArg_ParseTuple(args, "i:setAllowPinkMapping", &update);
    gps->allowPinkMapping = update;
    Py_XINCREF(Py_None);
    return Py_None;
}

static PyObject * VisionLink_getUsePinkUpdate(PyObject * /*self*/, PyObject *args) {
	return PyInt_FromLong(gps->pinkUpdate);
}

static PyObject * VisionLink_getTheCurrentMode(PyObject * /*self*/, PyObject *args) {
	return PyInt_FromLong(Behaviours::mode_);
}

static PyObject * VisionLink_getKickOffState(PyObject * /*self*/, PyObject *args) {
	return PyInt_FromLong(Behaviours::kickOff);
}

static PyObject * VisionLink_setMinimalMode(PyObject * /*self*/, PyObject *arg) {
    Behaviours::minimalModes = true;

    Py_XINCREF(Py_None);
    return Py_None;
}

//WARNING : duplicated with getBatteryLevel
static PyObject * VisionLink_getRemainingCapacity(PyObject * /*self*/, PyObject *args) {
#ifndef OFFLINE
	OStatus result;
	OPowerStatus currentStatus;
	result = OPENR::GetPowerStatus(&currentStatus);
	
	if (result == oSUCCESS) {
		return PyInt_FromLong(currentStatus.remainingCapacity);
	}
	else {
		// Return an invalid value.
		return PyInt_FromLong(-3);
	}
#endif // OFFLINE
}

static PyObject * VisionLink_getTemperature(PyObject * /*self*/, PyObject *args) {
#ifndef OFFLINE
	OStatus result;
	OPowerStatus currentStatus;
	result = OPENR::GetPowerStatus(&currentStatus);
	
	if (result == oSUCCESS) {
		return PyInt_FromLong(currentStatus.temperature);
	}
	else {
		// Return an invalid value.
		return PyInt_FromLong(-3);
	}
#endif // OFFLINE	
}

static PyObject * VisionLink_getCPUUsage(PyObject * /*self*/, PyObject *args) {
	return PyFloat_FromDouble(Behaviours::cpuUsage);
}

/* With this function specifying all the implemented semi-Python/semi-C
** functions, the python codes can call the "bridging" functions implemented in
** this file.
 *
 * NOTE: CHANGE IN THIS TABLE SHOULD BE TYPED IN VisionLink.py as well!
 **/
static PyMethodDef Vision_methods[] = {

    /*
     * General functions
     */
    {"setCallbacks", VisionLink_setCallbacks, METH_VARARGS, "Set the PyEmbed callbacks."},

    /*
     * Actuator related functions
     */ 
    {"sendAtomicAction", VisionLink_sendAtomicAction, METH_VARARGS, "Set the atomic action."},
    {"sendIndicators", VisionLink_sendIndicators, METH_VARARGS, "Set the LEDs, tail and ears."},
    {"sendMessageToBase", VisionLink_sendMessageToBase, METH_VARARGS, "Send wireless message back to base station, message is prefixed by '****'"},
    {"getBatteryLevel", VisionLink_getBatteryLevel, METH_NOARGS, "Get the robot's battery level."},
    {"getBatteryCurrent",VisionLink_getBatteryCurrent, METH_NOARGS, "Get the robots's battery current"},
    {"getFallenState", VisionLink_getFallenState, METH_NOARGS, "Have we fallen over?"},
    {"getPressSensorCount", VisionLink_getPressSensorCount, METH_O, "Get the state of the robot's touch sensors."},
    {"getJointSensor", VisionLink_getJointSensor, METH_VARARGS, "Get joint sensor from CommonSense."},
    {"getAnySensor", VisionLink_getAnySensor, METH_VARARGS, "Get joint sensor from CommonSense."},
    {"getJointPWMDuty", VisionLink_getJointPWMDuty, METH_VARARGS, "Get joint pwmDuty (or force sensor) from CommonSense."},
    {"getPWalkInfo", VisionLink_getPWalkInfo, METH_NOARGS, "getPWalkInfo() --> (walkType, step, PG, leadingLeg, stepComplete, desiredPan, desiredTilt, desiredCrane, isHighGain) ."},

    /*
     * Communication functions
     */
    {"sendEnvironmentTeamMates", VisionLink_sendEnvironmentTeamMates, METH_NOARGS, "Send our world model to our teammates.[not available]"},
    {"sendCompressedCPlane", VisionLink_sendCompressedCPlane, METH_NOARGS, "Send the CPlane to the base station.[not available]"},
    {"sendOPlane", VisionLink_sendOPlane, METH_NOARGS, "Send the OPlane to the base station.[not available]"},
    {"sendEnvironmentBaseStation", VisionLink_sendEnvironmentBaseStation, METH_NOARGS, "Send the world model to the base station.[not available]"},
    {"sendYUVPlane", VisionLink_sendYUVPlane, METH_NOARGS, "Send the YUV plane to the base station.[not available]"},
    {"sendMyBehaviourInfo", VisionLink_sendMyBehaviourInfo, METH_VARARGS, "sendMyBehaviourInfo( ballDist, isStriker, isGoalie, haveBall, kickedOff, isDefending, isBackingOff) --> send my behaviours information to our teammates, such as ballDist, isGoalie, isForward... ."},

    /*
     * GPS functions
     */
    {"setGPSGoal", VisionLink_setGPSGoal, METH_O, "Tell GPS which goal is ours."},
    {"setGPSPaused", VisionLink_setGPSPaused, METH_O, "Pause GPS updates.  Useful when penalised."},
    {"setIfGPSMotionUpdate", VisionLink_setIfGPSMotionUpdate, METH_VARARGS, "Set gps updating with motion."},
    {"setIfGPSVisionUpdate", VisionLink_setIfGPSVisionUpdate, METH_VARARGS, "Set gps updating with vision."},
    {"setIfGPSOpponentUpdate", VisionLink_setIfGPSOpponentUpdate, METH_VARARGS, "Set gps opponent updating."},
    {"resetEdgeMap", VisionLink_resetEdgeMap, METH_NOARGS, "Resets the edge pixel stationary map."},
    {"setAllowPinkMapping", VisionLink_setAllowPinkMapping, METH_VARARGS, "Set recording of pink blobs."},  
    {"setGlobalMaxEdge", VisionLink_setGlobalMaxEdge, METH_NOARGS, "Sets gps location for maximum edge match."},
    {"getUsePinkUpdate", VisionLink_getUsePinkUpdate, METH_NOARGS, "Did gps use a pink blob to map/localise off"},
    {"getIfGPSMotionUpdate", VisionLink_getIfGPSMotionUpdate, METH_NOARGS, "Is gps updating with motion?"},
    {"getIfGPSVisionUpdate", VisionLink_getIfGPSVisionUpdate, METH_NOARGS, "Is gps updating with vision?"},    
    {"getGPSSelfInfo", VisionLink_getGPSSelfInfo, METH_NOARGS, "Get our location info."},
    {"getGPSSelfCovariance", VisionLink_getGPSSelfCovariance, METH_NOARGS, "Get information about our own goal.\n return 3x3 tuple"},
    {"getGPSBallInfo", VisionLink_getGPSBallInfo, METH_VARARGS, "Get information on the ball."},
    {"getGPSBallMaxVar", VisionLink_getGPSBallMaxVar, METH_NOARGS, "see gps::getBallMaxVar() ."},
    {"getGPSBallVInfo", VisionLink_getGPSBallVInfo, METH_VARARGS, "Get information on the ball's velocity."},
    {"getGPSVBallMaxVar", VisionLink_getGPSVBallMaxVar, METH_NOARGS, "see gps::getVBallMaxVar() ."},
    {"getGPSOpponentInfo", VisionLink_getGPSOpponentInfo, METH_VARARGS, "Get information on an opponent.\n getGPSOpponentInfo(oppNum, context) ---> Vector where the opponent number oppNum (0..3) is, in context (GLOBAL,LOCAL)"},
    {"getGPSOpponentCovMax", VisionLink_getGPSOpponentCoxMax, METH_VARARGS, "Get covariance max of an opponent.\n getGPSOpponentCovMax(oppNum) --> double, oppNum = 0..3"},
    {"getGPSTeammateInfo", VisionLink_getGPSTeammateInfo, METH_VARARGS, "Get information about a teammates.\n getGPSTeammateInfo(teammate, context) --> Return WMObj tuple about that teammate [1..4] and context [LOCAL|GLOBAL] "},
    {"getOppGoalInfo", VisionLink_getOppGoalInfo, METH_NOARGS, "Get information about the opponent's goal."},
    {"getOwnGoalInfo", VisionLink_getOwnGoalInfo, METH_NOARGS, "Get information about our own goal."},
    {"getGPSCoordArray", VisionLink_getGPSCoordArray, METH_NOARGS, "Get information about our own goal."},
    {"getSelfLocation", VisionLink_getSelfLocation, METH_NOARGS, "Get self location.\n returns (confidence, x,y, heading) "},
    {"getBallLocation", VisionLink_getBallLocation, METH_NOARGS, "Get ball location in absolute coordinate(global).\n returns (confidence,x,y,distance) "},
    {"getTeamColor", VisionLink_getTeamColor, METH_NOARGS, "Get team color.\n returns 0 : blue, 1 : red "},
    {"getGPSTeammateBallInfo", VisionLink_getGPSTeammateBallInfo, METH_VARARGS, "Get ball info from teammate.\n getGPSTeammateBallInfo(tm) , tm = 0..3, returns pseudo WMObj"},

    /* 
     * Vision functions
     */ 
    {"getVisualObject", VisionLink_getVisualObject, METH_VARARGS, "Get visual information. index matches vob[] array "},
    {"getColouredPixelCount", VisionLink_dummy, METH_NOARGS, "Count the number of pixels of a given colour.[not available]"},
    {"getBallDistance", VisionLink_getBallDistance, METH_NOARGS, "Get distance from the ball to the robot. "},
    {"getBallHeading", VisionLink_getBallHeading, METH_NOARGS, "Get heading of the ball relative to the robot. "},
    {"getBallConfidence", VisionLink_getBallConfidence, METH_NOARGS, "Get confidence of the ball . "},
    {"setUseBeacons", VisionLink_setUseBeacons, METH_VARARGS, "Sets whether visualcortex looks for beacons and goals."},

    /* 
     * Wireless functions
     */ 
    {"getSharedBallInfo", VisionLink_getSharedBallInfo, METH_NOARGS, "Get information about the ball sent by our teammates."},
    {"getWirelessTeammateInfo", VisionLink_getWirelessTeammateInfo, METH_NOARGS, "Get information of wireless teammates."},
	
	/*
	 * GameController functions
	 */
    {"getTheCurrentMode", VisionLink_getTheCurrentMode, METH_NOARGS, "Get my current robot mode. Python should only receive mReadyMatch, mStartMatch or mNormal."},
    {"getKickOffState", VisionLink_getKickOffState, METH_NOARGS, "Get a boolean stating if my team is kicking off."},
    {"setMinimalMode", VisionLink_setMinimalMode, METH_NOARGS, "Sets minimal robot mode switching on, i.e. the pre-behvaiour stationary localising stage is skipped."},
	 
    /*
     * Other functions
     */
    {"doBasicBehaviour", VisionLink_doBasicBehaviour, METH_VARARGS, "Perform basic actions that are implemented using C++ codes in 2003."},
    {"getMyPlayerNum", VisionLink_getMyPlayerNum, METH_NOARGS, "Get my own player number."},
    {"getRemainingCapacity", VisionLink_getRemainingCapacity, METH_NOARGS, "Get the robot's remaining battery capacity."},
    {"getTemperature", VisionLink_getTemperature, METH_NOARGS, "Get the robot's temperature."},
    {"getCPUUsage", VisionLink_getCPUUsage, METH_NOARGS, "Get the cpu usage."},
    {"getCurrentPreviousFrameID", VisionLink_getCurrentPreviousFrameID, METH_NOARGS, "Get the current and previous OPENR frame ID."},
    {"getCurrentTime",VisionLink_getCurrentTime,METH_NOARGS,"getCurrentTime() -> (seconds, useconds). Get the system time"},

    /*
     * Some helper functions that are already availble in C++. We don't want to re-implement it in Python.
     */
    {"setWalkLearningParameters", VisionLink_setWalkLearningParameters, METH_VARARGS, "Send walk learining parameters to actuator control."},
    {"getProjectedBall", VisionLink_getProjectedBall, METH_NOARGS, "Get the ball project point."},
    {"getProjectedPoint", VisionLink_getProjectedPoint, METH_VARARGS, "Project a point to the ground plane."},
    {"getRedBlobInfo", VisionLink_getRedBlobInfo, METH_NOARGS, "."},
    {"getTestingInfo", VisionLink_getTestingInfo, METH_NOARGS, "."},
    {"printCPlusPlus", VisionLink_printCPlusPlus, METH_VARARGS, "Print C++ variables."},
    {"PointToElevation", VisionLink_PointToElevation, METH_VARARGS, "."},
    {"setCPlaneMode", VisionLink_setCPlaneMode, METH_VARARGS, "."},
    {"PointToHeading", VisionLink_PointToHeading, METH_VARARGS, "."},
    {"getHeadingToBestGap", VisionLink_getHeadingToBestGap, METH_NOARGS, "Get the best heading to the goal considering obstacles, see voak."},
    {NULL,    NULL, 0, NULL}    /* sentinel */
};

static PyObject * initVisionLinkModule(void) {
    PyObject *module;
    if ((module = PyImport_AddModule(VisionLinkModuleName)) == NULL) {
        cout << "error creating VisionLink Module" << endl;
        return NULL;
    }
    else if (Py_InitModule(VisionLinkModuleName, Vision_methods) == NULL) {
        cout << "error linking VisionLink Module" << endl;
        return NULL;
    }
    return module;
}
