/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/



/*
 * Last modification background information
 * $Id: gps.cc 
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 *
**/

#include <cmath>
#include <cstdio>
#include <iostream>

#ifdef VALGRIND
#include <valgrind/memcheck.h>
#endif

#include "gps.h"
#include "Possible.h"
#include "KalmanInfo2D.h"
#include "KI2DWithVelocity.h"
#include "../share/Common.h"
#include "../share/SharedMemoryDef.h"
#include "../share/WirelessSwitchboard.h"

using namespace std;

#define ERICDEBUG 0

static const bool goalSkippedSilence = true;
static const int TEAMMATE_GPS_COUNTER = 8;

static const bool useSymmetry = false;
static const double symmetryWeight = 0.05;

// Possible coordinates for the vobs
static const long coordRedBased[8][2] =
{
    {FIELD_WIDTH / 2, FIELD_LENGTH + WALL_THICKNESS},   //far goal
    {FIELD_WIDTH / 2, -WALL_THICKNESS},                 //close goal

    {0, FIELD_LENGTH},                                  //far left beacon
    {FIELD_WIDTH, FIELD_LENGTH},                        //far right beacon
    {0, FIELD_LENGTH / 2},                              //middle left beacon
    {FIELD_WIDTH, FIELD_LENGTH / 2},                    //middle right beacon
    {0, 0},                                             //close left beacon
    {FIELD_WIDTH, 0}                                    //close right beacon
};

static const long coordBlueBased[8][2] =
{
    {FIELD_WIDTH / 2, -WALL_THICKNESS},                 //close goal
    {FIELD_WIDTH / 2, FIELD_LENGTH + WALL_THICKNESS},   //far goal

    {FIELD_WIDTH, 0},                                   //close right beacon
    {0, 0},                                             //close left beacon
    {FIELD_WIDTH, FIELD_LENGTH / 2},                    //middle right beacon
    {0, FIELD_LENGTH / 2},                              //middle left beacon
    {FIELD_WIDTH, FIELD_LENGTH},                        //far right beacon
    {0, FIELD_LENGTH}                                   //far left beacon
};

// for opponent tracking
static const double oppIniX[4] = 
        {FIELD_WIDTH/2, FIELD_WIDTH/2, 3*FIELD_WIDTH/4, FIELD_WIDTH/2};
static const double oppIniY[4] = 
        {FIELD_LENGTH, 3*FIELD_LENGTH/4 , 3*FIELD_LENGTH/4, FIELD_LENGTH/2};
static const double friendIniX[NUM_FRIENDS_VIS] = 
        {FIELD_WIDTH/2, FIELD_WIDTH/2, 3*FIELD_WIDTH/4/*, FIELD_WIDTH/2*/};
static const double friendIniY[NUM_FRIENDS_VIS] = 
        {WALL_THICKNESS,FIELD_LENGTH/4, FIELD_LENGTH/4/*, FIELD_LENGTH/2*/};
static const double oppIniXVar = get95CF(100);
static const double oppIniYVar = get95CF(100);

static const bool raySilence = true;

// maximum dimension of a "measurement" for the frame
static const int MAX_MEASUREMENT_DIM = 9;

// the growth in opponent position variance
static const double OPPONENT_VAR_GROWTH = get95CF(2);

// varience in movement = P
static const int MOTION_VAR = 10;
static const int MOTION_H_VAR = 6;
static const int WIRELESS_VAR = 50;

static const int vobMinBlueDog = vobBlueDog;
static const int vobMaxBlueDog = vobBlueDog4;
static const int vobMinRedDog = vobRedDog;
static const int vobMaxRedDog = vobRedDog4;

// These determine how small a gaussian component of r has to be before
//  it is culled from the sum, and how close together two components
//  can be before they are merged.
static const double cullThreshold = 1E-50;
static const double mergeThreshold = 20.0;
static const double mergeCovThreshold = 10000.0;
    
// This determines at what probability a measurement is regarded as
//  noise, and so will be ignored
static const double noiseThreshold = 1E-2;

// weight for exponential decay of visPull
static const int vpHistoryLength = 25;
//static const double visPullWeight = 0.8;

static const double badVisionUpdateWeight = 0;	// 0.01;	// probability that a vision update is bad

static const bool useRelativeCullWeight = true;

GPS::GPS () {
    targetGoal = vobBlueGoal;    // by default the robot is going towards blue goal
    goalNotSet = true;

    frameId = 0;
	selfAccessId = -1;
    ballAccessId = -1;
	lastBallOrigin = 'n';
    teammateAccessId = -1;
    lastTeammateAccessed = -1;
    lastTeammateOrigin = 'n';
	oGoalAccessId = -1;
    tGoalAccessId = -1;
    vBallAccessId = -1;
	lastvBallOrigin = 'n';

	// Update or not?
	doGPSMotionUpdate = true;
	doGPSVisionUpdate = true;
    doGPSOpponentUpdate = true;

    //!!ERIC these are temp for distance calibration
    sendDebug = false;
    debugCounter = 0;
    debugParm = 0;
	
	isPenalized = false;

    double rx = FIELD_WIDTH / 2.0;
    double ry = FIELD_LENGTH / 2.0;
    	
	// Robot position probability starts with only one gaussian
    numGaussians = 1;
    r[0].mean.pos.setVector(vCART,rx,ry);
    r[0].mean.h = HALF_CIRCLE/2;
    r[0].cov(0,0)= get95CF (FIELD_WIDTH);
    r[0].cov(0,1)= 0.0;
    r[0].cov(0,2)= 0.0;
    r[0].cov(1,0)= 0.0;
    r[0].cov(1,1)= get95CF (FIELD_LENGTH);
    r[0].cov(1,2)= 0.0;
    r[0].cov(2,0)= 0.0;
    r[0].cov(2,1)= 0.0;
    r[0].cov(2,2)= get95CF (HALF_CIRCLE);
	r[0].weight = 1.0;
	r[0].mean.updateVar(&r[0].cov);
	
	vbPull(0,0) = 0.0;
	vbPull(1,0) = 0.0;
	vbPull(2,0) = 0.0;
	
	visPull(0,0) = 0.0;
	visPull(1,0) = 0.0;
	visPull(2,0) = 0.0;
	vpHistory = new MVec3[vpHistoryLength];
	for(int i=0 ; i<vpHistoryLength ; i++)
		vpHistory[i].reset();
	vpIndex = 0;

    unitMatrix3 = new MMatrix3(1);
    unitMatrix4 = new MMatrix4(1);

    noBallFrames = 1;

    //initialise all teammates to not exist
    //when you get a message from them over ride this
    for (int i=0; i<NUM_TEAM_MEMBER; i++) {
		wmTeammate[i].hVar=INVALID_FLAG;
	}

	for (int i=0; i<NUM_TEAM_MEMBER; i++) {
		tmBall[i].pos.setVector(vCART,0,0);
		tmBall[i].posVar = get95CF(500);
	}

    // initialise noone can see ball
    shareBallx = 0;
    shareBally = 0;
    shareBallvar = LARGE_VAL;
    sbRobotNum = -1;
	
	newball = new KI2DWithVelocity(this, FIELD_WIDTH/2, FIELD_LENGTH/2, get95CF(FIELD_WIDTH/2), get95CF(FIELD_LENGTH/2),
									0.0, 0.0, get95CF(100), get95CF(100));
    
    // initialise opponent robots' positions
    for(int i=0 ; i<NUM_OPPONENTS ; i++)
        opponents[i] = new KalmanInfo2D(this,oppIniX[i],oppIniY[i],oppIniXVar,oppIniYVar);

#ifdef VISUAL_TRACK_FRIENDS
    for(int i=0 ; i<NUM_FRIENDS_VIS ; i++)
        friends[i] = new KalmanInfo2D(this,friendIniX[i],friendIniY[i],oppIniXVar,oppIniYVar);
#endif // VISUAL_TRACK_FRIENDS

    view = 0;
    
    directx = -1.0;
    directy = -1.0;
    directh = -1.0;

    motionForward = 0.0;
    motionLeft = 0.0;
    motionTurn = 0.0;
    
	doLatTest = false;
	
	oldx = r[0].mean.pos.x;
    oldy = r[0].mean.pos.y;
    oldh = r[0].mean.h;
    oldvx = r[0].cov(0,0);
    oldvy = r[0].cov(1,1);
    oldvh = r[0].cov(2,2);
    oldcxy = r[0].cov(0,1);
    oldcxh = r[0].cov(0,2);
    oldcyh = r[0].cov(1,2);
    
    edgeDebugCounter = 0;
    
    // For localisation challenge
    numPink = 0;
    pinkDecreaseCounter = 0;
    pinkUpdate = false;
    allowPinkMapping = false;
    landmarksInitialised = false;
    for (int i = 0; i < maxPink; i++)
        pinkConf[i] = 0;
    
    // For weight scaler statistics
    for (int i = 0; i < 10; i++) {
        nX[i] = 0;
        sumX[i] = 0.0;
        sumXSquared[i] = 0.0;
    }
    
    //testKalman(); // DERRICK DEBUG
}

// append gps info
int GPS::storeGps(unsigned char *add, int base) {
    int ret = base;
    
	/*
    memcpy(add + ret, &numGaussians, sizeof(numGaussians));
	ret += sizeof(numGaussians);
    
    for (int i = 0; i < numGaussians; i++) {
	    ret = r[i].mean.serialize(add, ret);
    	ret = r[i].cov.serialize(add, ret);
		memcpy(add + ret, &r[i].weight, sizeof(double));
		ret += sizeof(double);
	}*/
    // Only store first gaussian for the moment (this should be temporary)
    ret = r[0].mean.serialize(add, ret);
    ret = r[0].cov.serialize(add, ret);
		
    ret = visPull.serialize(add, ret);
    ret = newball->serialize(add, ret);
    ret = vbPull.serialize(add, ret);
    
    for(int i=0 ; i<NUM_TEAM_MEMBER ; i++)
        ret = wmTeammate[i].serialize(add, ret);
    for(int i=0 ; i<NUM_OPPONENTS ; i++)
        ret = opponents[i]->serialize(add, ret);
    
	return ret;
}

// restore gps information
int GPS::restoreGps(unsigned char *add, int base) {
    
    int ret = base;
	
    /*
    memcpy(&numGaussians, add + ret, sizeof(numGaussians));
	ret += sizeof(numGaussians);
    
	for (int i = 0; i < numGaussians; i++) {
	    ret = r[i].mean.unserialize(add, ret);
    	ret = r[i].cov.unserialize(add, ret);
		memcpy(&r[i].weight, add + ret, sizeof(double));
		ret += sizeof(double);
	}*/
    
    // Only restore first gaussian for the moment (this should be temporary)
    numGaussians = 1;
    ret = r[0].mean.unserialize(add, ret);
    ret = r[0].cov.unserialize(add, ret);    
    r[0].weight = 1.0;
	r[0].mean.updateVar(&r[0].cov);
    
    ret = visPull.unserialize(add, ret);
    ret = newball->unserialize(add, ret);
    ret = vbPull.unserialize(add, ret);
    
    for(int i=0 ; i<NUM_TEAM_MEMBER ; i++)
        ret = wmTeammate[i].unserialize(add, ret);
    for(int i=0 ; i<NUM_OPPONENTS ; i++)
        ret = opponents[i]->unserialize(add, ret);
	
	return ret;
}

//This is called by Vision.cc
//to set your goals and orientation of the beacons.
void GPS::SetGoals(bool isRedRobot) {
    if (!isRedRobot) {
        targetGoal        = vobYellowGoal;
        ownGoal           = vobBlueGoal;
        backLeftBeacon    = vobBlueRightBeacon;
        backRightBeacon   = vobBlueLeftBeacon;
        middleLeftBeacon  = vobGreenRightBeacon;
        middleRightBeacon = vobGreenLeftBeacon;
        farLeftBeacon     = vobYellowRightBeacon;
        farRightBeacon    = vobYellowLeftBeacon;

        coord              = &coordBlueBased;

		for (int i = 0; i < maxGaussians; i++)
        	r[i].mean.vobType = 13; //(first vob number for a red dog in the vob list)
        
        vobMinOpponent = vobMinRedDog;
        vobMaxOpponent = vobMaxRedDog;
#ifdef VISUAL_TRACK_FRIENDS
        vobMinFriend = vobMinBlueDog;
        vobMaxFriend = vobMaxBlueDog;
#endif // VISUAL_TRACK_FRIENDS

    } else {
        targetGoal        = vobBlueGoal;
        ownGoal           = vobYellowGoal;
        backLeftBeacon    = vobYellowLeftBeacon;
        backRightBeacon   = vobYellowRightBeacon;
        middleLeftBeacon  = vobGreenLeftBeacon;
        middleRightBeacon = vobGreenRightBeacon;
        farLeftBeacon     = vobBlueLeftBeacon;
        farRightBeacon    = vobBlueRightBeacon;

        coord              = &coordRedBased;

        for (int i = 0; i < maxGaussians; i++)
        	r[i].mean.vobType = 9; //(first vob number for a blue dog in the vob list)
        
        vobMinOpponent = vobMinBlueDog;
        vobMaxOpponent = vobMaxBlueDog;
#ifdef VISUAL_TRACK_FRIENDS
        vobMinFriend = vobMinRedDog;
        vobMaxFriend = vobMaxRedDog;
#endif // VISUAL_TRACK_FRIENDS
        
    }
    goalNotSet = false;
}

void GPS::setPenalized(bool p) {
	isPenalized = p;
}

// Next few methods are used for the localisation challenge.
// The next two are called by ResultCamera in Vision.cc.

// This calculates where the given pink blobs would be if they were attached
// to the perimeter walls, and records that position.

static const double mergeDistSq = SQUARE(25.0);

void GPS::MapLocalisePink(VisualObject *pvo, int numPinkVob) {
    double x, y, dx, dy, distSq, scale;
    bool merged;
    int i, j;
    
    pinkUpdate = false;

    if (!allowPinkMapping) return;
    
    for (i = 0; i < numPinkVob; i++) {
        if (CalcWallIntercept(&pvo[i], &x, &y)) {
            merged = false;
            pinkUpdate = true;
            
            // For every valid pink blob, see if it can be merged with one
            // of the mapped pink blobs, by checking distance from each.
            for (j = 0; j < numPink && (!merged); j++) {
                dx = x - pinkX[j]; dy = y - pinkY[j];
                distSq = dx * dx + dy * dy;
                
                // If this pink blob is close enough to be merged with another,
                // then use it in the running average.
                if (merged = (distSq < mergeDistSq)) {
                    scale = 1.0 / (1 + pinkConf[j]);
                    pinkX[j] += dx * scale;
                    pinkY[j] += dy * scale;
                    pinkConf[j]++; // Increase confidence value
                }
            }

            // If the blob couldn't be merged with another, start a new one.            
            if (!merged && numPink < maxPink) {
                pinkX[numPink] = x;
                pinkY[numPink] = y;
                pinkConf[numPink] = 1;
                numPink++;
            }
        }
    }
    
    // Slowly decrease confidence (to hopefully eliminate pink noise)
    pinkDecreaseCounter++;
    if (pinkDecreaseCounter > 120) {
        ReducePinkConfidence(3, 1);
        pinkDecreaseCounter = 0;
    }
}

// This uses the largest valid pink blob seen to localise with the pink blobs
// recorded by MapLocalisePink. Only an angle update is performed.

void GPS::LocalisePinkUpdate(VisualObject *pvo, int numPinkVob) {
    double jacX, jacY, jacH, iv, var, minWeight;
    int oldNumGaussians, mode, obsMode, dummy;
    int i, vobArea, useVob;
    SingleGaussian temp;
    SlimVisOb obs;
    
    pinkUpdate = false;

    if (!landmarksInitialised) {
        // Cull any low confidence pink blobs
        ReducePinkConfidence(3, 3);
    
        DER_DEBUG(
            cout << "Found " << numPink << " pink landmarks.\n";

            for (i = 0; i < numPink; i++) {
                cout << "Landmark " << i;
                cout << " - x: " << pinkX[i] << "   y: " << pinkY[i];
                cout << "   conf: " << pinkConf[i] << "\n";
            }
        );
    
        landmarksInitialised = true;
    }

	if (!doGPSVisionUpdate)
		return;

    // Find the largest valid pink blob
    vobArea = 0; useVob = -1;
    for (i = 0; i < numPinkVob; i++) {
        // Do an elevation check on the blob
        if (pvo->elev < 10.0 && pvo->elev > -2.0) {

            // If this is the largest pink blob, use it
            if (useVob < 0 || pvo->area > vobArea) {
                useVob = i;
                vobArea = pvo->area;
            }
        }
    }
    
    // If we found a suitable blob, angle localise off it
    pinkUpdate = (useVob >= 0);

    if (pinkUpdate) {
        // Work from a backup of the old distribution
        minWeight = r[numGaussians - 1].weight;
        oldNumGaussians = numGaussians;
        for (i = 0; i < numGaussians; i++) {
            rCopy[i] = r[i];
            r[i].weight *= noiseThreshold;
        }
        	
        mode = 0;    
        while (mode < oldNumGaussians && rCopy[mode].weight >= minWeight) {
            for (obsMode = 0; obsMode < numPink; obsMode++) {
        
                // Calculate the effect of every observation mode on each of
                // the distribution modes
                dummy = 0;
                temp = rCopy[mode];
                obs.h = pvo[useVob].head;
                obs.angleVar = 400.0;
                addMeasurement(&temp, obs, pinkX[obsMode], pinkY[obsMode],
                    &jacX, &jacY, &jacH, &iv, &var, dummy, false);
            
                kalmanUpdate<1>(&temp, &jacX, &jacY, &jacH, &iv, &var);
            
                // Insert the corrected mode into the new distribution
                insertGaussian(temp);
                if (numGaussians == maxGaussians)
                    minWeight = r[numGaussians - 1].weight;
            }

            mode++;
        }
        
        formDistribution();
    }
}

// A set of 4 (x, y) gps coordinates that specify the perimeter wall corner
// coordinates. Should to be in clockwise order.

static double perimeter[8] = {-109, -109,
                              -111, FIELD_LENGTH + 110,
                              FIELD_WIDTH + 108, FIELD_LENGTH + 108,
                              FIELD_WIDTH + 109, -108};

static double beaconPos[8] = {0, 0,
                              0, FIELD_LENGTH,
                              FIELD_WIDTH, FIELD_LENGTH,
                              FIELD_WIDTH, 0};
                              
// Given a visual object, this takes the heading and elevation and determines
// the intercept with the perimeter wall, returned in iX, iY. Returns true
// if the VisualObject should be accepted as a legitimate pink blob.

bool GPS::CalcWallIntercept(VisualObject *pvo, double *iX, double *iY) {
    double wallDX, wallDY, wallA, wallB, wallC;
    double rayHead, rayX, rayY, dist, dotProd;
    double intersectX, intersectY;
    int curPoint, nextPoint;

    // We should make sure self variance is small if we are going to record
    // the position of this blob.
    if (r[0].mean.posVar > 500.0 || r[0].mean.hVar > 300.0)
        return false;
    
    // Calculate ray direction parameters
    rayHead = r[0].mean.h + pvo->head;
    rayX = cos(DEG2RAD(rayHead));
    rayY = sin(DEG2RAD(rayHead));

    // Make sure ray doesn't intersect a beacon position
    for (curPoint = 0; curPoint < 8; curPoint += 2) {
        // Make sure beacon is in ray direction
        intersectX = beaconPos[curPoint] - r[0].mean.pos.x;
        intersectY = beaconPos[curPoint + 1] - r[0].mean.pos.y;

        if (intersectX * rayX + intersectY * rayY >= 0) {
            // If perpendicular distance from ray to beacon is less than a
            // threshold value, we might have accidentally taken the pink blob
            // from a beacon, so don't use it.
            dotProd = intersectX * rayY - intersectY * rayX;
            if (fabs(dotProd) < 25) return false;
        }
    }
    
    curPoint = 6;
    for (nextPoint = 0; nextPoint < 8; nextPoint += 2) {
        // Check for an intersection between each wall and a ray from robot
        // position, in the given vob's direction
        wallDX = perimeter[nextPoint] - perimeter[curPoint];
        wallDY = perimeter[nextPoint + 1] - perimeter[curPoint + 1];

        // Calculate wall line paramters
        wallA = wallDY; wallB = -wallDX;
        wallC = perimeter[nextPoint] * perimeter[curPoint + 1];
        wallC -= perimeter[curPoint] * perimeter[nextPoint + 1];
        
        // Find intersection between heading ray and wall line
        dist = -(wallA * r[0].mean.pos.x + wallB * r[0].mean.pos.y + wallC);
        dist /= (wallA * rayX + wallB * rayY);

        // Check that the elevation of the blob is not too high/low.
        if (pvo->elev > (5.0 + 500.0 / dist) || pvo->elev < -2.0)
            return false;
        
        if (dist >= 0) {
            // Make sure intersection occurs within wall interval
            intersectX = rayX * dist + r[0].mean.pos.x;
            intersectY = rayY * dist + r[0].mean.pos.y;
            
            dotProd = (intersectX - perimeter[curPoint]) * wallDX;
            dotProd += (intersectY - perimeter[curPoint + 1]) * wallDY;
            dist = wallDX * wallDX + wallDY * wallDY;

            if (dotProd >= 0 && dotProd < dist) {
                (*iX) = intersectX;
                (*iY) = intersectY;
                return true;
            }            
        }

        curPoint = nextPoint;
    }

    return true;
}

void GPS::ReducePinkConfidence(int minConf, int decreaseAmount) {
    int newNumPink = 0;
    bool deleted;
    
    for (int i = 0; i < numPink; i++) {
        // Only reduce confidence if it is low to start with
        if (pinkConf[i] <= minConf) {
            pinkConf[i] -= decreaseAmount;

            // Delete any modes whos confidence drops to zero
            deleted = (pinkConf[i] <= 0);

        } else deleted = false;

        // Compact list of pink modes if any modes have been deleted
        if (!deleted) {
            if (newNumPink < i) {
                pinkX[newNumPink] = pinkX[i];
                pinkY[newNumPink] = pinkY[i];
                pinkConf[newNumPink] = pinkConf[i];
            }
        
            newNumPink++;
        }
    }
        
    numPink = newNumPink;
}

//gets called by ReadyEffectors in ActuatorControl
//which means it gets called everytime a pwalk half step is completed
//
// This is the time update part of the Kalman filter
// Takes in feedback from locomotion and predicts a new location
//
// Calculates a priori state
// dForward - amount moved forward in cm
// dLeft - amount moved to the left in cm
// dTurn - degrees turned to the left from straight ahead
// note - assumes that turn is performed last
//

int MotionUpdateCount = 0; //for debugging only

void GPS::GPSMotionUpdate(double dForward, double dLeft, double dTurn, long PWM) {

	if (!doGPSMotionUpdate)
		return;

	if(isPenalized)
		return;
    
    // Check for NaN or Inf inputs
    if (dForward != dForward || dLeft != dLeft || dTurn != dTurn) {
        cout << "GPS::GPSMotionUpdate error: One or more inputs are NaN or Inf!\n";
        return;
    }

    motionForward += fabs(dForward);
    motionLeft += fabs(dLeft);
    motionTurn += fabs(dTurn);

    // Increase variance faster if we have detected that we are stuck.
    double stuckVariance, stuckScaler;
    if (PWM < 900) stuckVariance = 0.0;
    else stuckVariance = (PWM - 300) * 0.05;
    
    stuckScaler = 1.0;
    DER_DEBUG(
        // Don't use full motion update if stuck
        if (PWM > 750) {
            stuckScaler = 1.0 / (1.0 + SQUARE(PWM - 750) * 0.0001);
            cout << "Stuck scaler: " << stuckScaler << "\n";
        }
    )
  
	for (int i = 0; i < numGaussians; i++) {
		double dx, dy, dh;
		
	    // determine movement according to our original heading
	    dx = dForward * cos (DEG2RAD(r[i].mean.h));
		dx -= dLeft * sin (DEG2RAD(r[i].mean.h));
	    dy = dForward * sin (DEG2RAD(r[i].mean.h));
		dy += dLeft * cos (DEG2RAD(r[i].mean.h));
    	dh = dTurn;

        dx *= stuckScaler;
        dy *= stuckScaler;
        dh *= stuckScaler;

	    // save original heading and then alter it
	    double oh = r[i].mean.h;
    	r[i].mean.h += dh;

		// alter dx and dy due to the difference between the turn centre and the
    	// centre of localisation (neckbase) - this is directly from old code 2002
	    dx += TURN_CENTRE_TO_NECK * (cos(DEG2RAD(r[i].mean.h)) - cos(DEG2RAD(oh)));
    	dy += TURN_CENTRE_TO_NECK * (sin(DEG2RAD(r[i].mean.h)) - sin(DEG2RAD(oh)));

	    // update current position and clip (to be added)
    	double rx = r[i].mean.pos.x + dx;
	    double ry = r[i].mean.pos.y + dy;
	    ANGLE_FIX(r[i].mean.h);
    	FieldClipping(rx, ry);
	    r[i].mean.pos.setVector(vCART,rx,ry);

	    // update variance
		r[i].cov(0,0) += 2 + get95CF(dx) + stuckVariance;
	    r[i].cov(1,1) += 2 + get95CF(dy) + stuckVariance;
    	r[i].cov(2,2) += 0.3 + get95CF(dh) + stuckVariance * 0.5;
	
	    r[i].mean.updateVar(&r[i].cov);
	}
}

// This generally gets called when robot is penalized and it's position is
// to be reset
void GPS::GPSSetVals(double newx, double newy, double newh, const MMatrix3& newC) {
    // Check for NaN inputs
    if (newx != newx || newy != newy || newh != newh) {
        cout << "GPS::GPSSetVals error: One or more inputs are NaN or Inf!\n";
        return;
    }
    
	numGaussians = 1;
    
	r[0].mean.pos.setVector(vCART,newx,newy);
	r[0].mean.h = newh;
	r[0].cov = newC;
	r[0].weight = 1.0;	
	r[0].mean.updateVar(&r[0].cov);
}

// Debugging method that outputs current multi-modal distribution data

void GPS::GPSOutputSelfData() {
    cout << "=== GPSOutputSelfData: " << numGaussians << " modes ===\n";

    for (int i = 0; i < numGaussians; i++) {
        cout << "Gaussian " << i << " has weight " << r[i].weight << "\n";
        cout << "Mean: x " << r[i].mean.pos.x << "   y " << r[i].mean.pos.y;
        cout << "   h " << r[i].mean.h << "   Covariance:\n";
        r[i].cov.printOut();
    }
    
    // Now output update stastics
    for (int i = 0; i < 10; i++) {
        double mean, stdDev;
    
        if (nX[i] > 0) {
            mean = sumX[i] / nX[i];
            stdDev = sqrt(sumXSquared[i] - SQUARE(mean));
        } else {
            mean = 0.0;
            stdDev = 0.0;
        }
    
        cout << nX[i] << " " << i << " dimensional updates. ";
        cout << "Mean: " << mean << "   Standard dev: " << stdDev << "\n";
    }
}

// Gets called by ResultCamera in Vision.cc
// which means it gets called on every camera frame?
// it is the measurement correction part of the kalman filter.
//
// This is the Measurement update part of the Kalman filter
// Takes in an observation from vision and "corrects" the time update predition
//
// Calculates a posteriori state
// nvo - an array of all types of VisualObjects
// On every camera frame vision sets this array (ie sets distances and variances
// to all the types of vobs)
// If a vob was not seen in that frame then its variance is set high

void GPS::GPSVisionUpdate (VisualObject * nvo, int ballDistMeth) {
    SingleGaussian temp;
    double lastWeight = 0.0;

	frameId++;

	if (!doGPSVisionUpdate)
		return;
	
	// Prevent the frameId growing too large..
	// in a 20 minute game, the id should grow anywhere close to this though.
	if (frameId > 1000000)
		frameId=0;
	
	// Reset behaviours value.		
    for (int i = 0; i < numGaussians; i++)
        r[i].mean.behavioursVal = 0;
	
	// First process friendly teammates, throwing out info that is too old.
	for(int i=0 ; i < NUM_TEAM_MEMBER ; i++) {
		if(wmTeammate[i].counter > 0) {
			wmTeammate[i].counter--;
			if((wmTeammate[i].counter == 0) && (sbRobotNum == (i+1)))
				shareBallvar = LARGE_VAL;
		}
	}


    //cout << "vision update" << endl;
    UpdateObservedVobs(nvo);
    view = 0;
    
    directx = -1.0;
    directy = -1.0;
    directh = -1.0;
	
	if(isPenalized)
		return;

	// ross's debug stuff - keep record of old thoughts for Robocommander
/*    oldx = r.pos.x;
    oldy = r.pos.y;
    oldh = r.h;
    oldvx = (*rcov)(0,0);
    oldvy = (*rcov)(1,1);
    oldvh = (*rcov)(2,2);
    oldcxy = (*rcov)(0,1);
    oldcxh = (*rcov)(0,2);
    oldcyh = (*rcov)(1,2);*/
	
	double jacX[MAX_MEASUREMENT_DIM];
	double jacY[MAX_MEASUREMENT_DIM];
	double jacH[MAX_MEASUREMENT_DIM];
	double var[MAX_MEASUREMENT_DIM];
	double iv[MAX_MEASUREMENT_DIM];
	int measDim, mode;

    // Make a low probability, large variance mode
    if (numGaussians < maxGaussians) numGaussians++;
    MMatrix3 largeCov(get95CF(500));
    largeCov(2, 2) = get95CF(300);
    r[numGaussians - 1].mean.pos.setVector(vCART, FIELD_WIDTH / 2, FIELD_LENGTH / 2);
    r[numGaussians - 1].mean.h = HALF_CIRCLE / 2;
    r[numGaussians - 1].cov = largeCov;
    r[numGaussians - 1].weight = 1E-12;
    r[numGaussians - 1].mean.updateVar(&r[numGaussians - 1].cov);
        
	// copy the current modes
	// with a low weight
	// to account for the possibility of bad vision
    int oldNumGaussians = numGaussians;
    for (mode = 0; mode < numGaussians; mode++) {
        rCopy[mode] = r[mode];
        r[mode].weight *= noiseThreshold;
    }
    
    int maxSymmetry = 1;
    if (useSymmetry) maxSymmetry = 2;
    
    for (mode = 0; mode < oldNumGaussians; mode++) {
        for (int symmetry = 0; symmetry < maxSymmetry; symmetry++) {
            for (int i = 0; i < MAX_MEASUREMENT_DIM; i++) {
                jacX[i] = 0.0;
                jacY[i] = 0.0;
                jacH[i] = 0.0;
                var[i] = 0.0;
                iv[i] = 0.0;
            }
            measDim = 0;

            temp = rCopy[mode];
            if (symmetry == 1) {
                temp.mean.pos.setVector(vCART,
                    FIELD_WIDTH - temp.mean.pos.x,
                    FIELD_LENGTH - temp.mean.pos.y);
                temp.mean.h = NormalizeAngle_0_360(temp.mean.h + 180);
            }

            bool skipGoal, useDist;
            double cx = temp.mean.pos.x;
            double cy = temp.mean.pos.y;

            for (int i = vobMinFixed; i < vobMinFixed + 2; i++) {
                // Don't localise off goal if angular variance is too big, or we're
                // too close (heading and distance accuracy can be bad).
                skipGoal = (observedVobs[i].angleVar > get95CF(30));
                skipGoal |= (observedVobs[i].d < 70);

                // Don't localise off goal if we know we're in a position where
                // the goal heading is going to be bad.
                skipGoal |= (i == ownGoal && cy < FIELD_LENGTH * 0.15);
                skipGoal |= (i == targetGoal && cy > FIELD_LENGTH * 0.85);

                // Don't localise off goal if we can't fit any more measurements in
                skipGoal |= (measDim >= MAX_MEASUREMENT_DIM);

                if(!goalSkippedSilence && skipGoal)
                    cout << "goal skipped" << endl;

                if (!skipGoal) {
                    view |= 1 << (i - vobMinFixed + 1);
    
                    useDist = (observedVobs[i].var < get95CF(100));
                    useDist &= (observedVobs[i].d < 220.0);
                    useDist &= (measDim < MAX_MEASUREMENT_DIM - 1);
                    
                    addMeasurement(&temp, observedVobs[i], jacX, jacY, jacH,
                                   iv, var, measDim, useDist);
                                 
                    if (useDist) measDim += 2;
                    else measDim++;
                
                }
            }

            //for all beacons where your estimate of the distance to the
            //beacon has a 95% confidence interval of less than 100cm
            //localise using that beacon
            for (int i = vobMinFixed + 2; i <= vobMaxFixed; i++) {
                if (observedVobs[i].angleVar < get95CF(30)) {                
                    view |= 1 << (i - vobMinFixed + 1);               

                    useDist = (observedVobs[i].var < get95CF(100));
                    useDist &= (observedVobs[i].d < 450);
                    useDist &= (measDim < MAX_MEASUREMENT_DIM - 1);
                
                    if(measDim < MAX_MEASUREMENT_DIM) {
                        addMeasurement(&temp, observedVobs[i], jacX, jacY, jacH,
                                       iv, var, measDim, useDist);
                                   
                        if (useDist) measDim += 2;
                        else measDim++;
                    }
                }
            }
        
            switch(measDim) {
            case 0: break;
            case 1: kalmanUpdate<1>(&temp, jacX, jacY, jacH, iv, var); break;
            case 2: kalmanUpdate<2>(&temp, jacX, jacY, jacH, iv, var); break;
            case 3: kalmanUpdate<3>(&temp, jacX, jacY, jacH, iv, var); break;
            case 4: kalmanUpdate<4>(&temp, jacX, jacY, jacH, iv, var); break;
            case 5: kalmanUpdate<5>(&temp, jacX, jacY, jacH, iv, var); break;
            case 6: kalmanUpdate<6>(&temp, jacX, jacY, jacH, iv, var); break;
            case 7: kalmanUpdate<7>(&temp, jacX, jacY, jacH, iv, var); break;
            case 8: kalmanUpdate<8>(&temp, jacX, jacY, jacH, iv, var); break;
            case 9: kalmanUpdate<9>(&temp, jacX, jacY, jacH, iv, var); break;
            }

            if (symmetry == 0) {
                // Insert the updated mode into the new distribution
                insertGaussian(temp);
                lastWeight = temp.weight;
            } else {
                // For the symmetrically opposite mode, only insert if it
                // matched up much better than the normal update, and matched
                // up well in general.
                double weightSymmetric = temp.weight / lastWeight;
                double weightScale = temp.weight / rCopy[mode].weight;

                if (weightSymmetric > 100 && weightScale > 0.9 && measDim > 1) {
                    DER_DEBUG (
                        cout << "Introduced symmetric mode\n";
                        cout << "Weight compared to normal mode: " << weightSymmetric;
                        cout << "\nWeight compared to original: " << weightScale;
                        cout << "\nUpdate dimensions: " << measDim << "\n";
                        sendCPlane = true;
                    );
                    temp.weight = lastWeight * symmetryWeight;
                    insertGaussian(temp);
                }
            } 
        }
    }

    formDistribution();

    newball->growVariance(4, 5);
    if(observedVobs[vobBall].cf > 0) {
        newball->infoKalman(observedVobs[vobBall], ballDistMeth);
        view |= 1;
    }
	
#ifdef VISUAL_TRACK_FRIENDS
    // perform prediction update on each friend structure
    for(int i = 0; i < NUM_FRIENDS_VIS; i++)
        friends[i]->growVariance(OPPONENT_VAR_GROWTH);

    // perform correction update for each friend seen
    for(int i = vobMinFriend; i <= vobMaxFriend; i++) {
        if(observedVobs[i].var < get95CF(150))
            //applyKalmanUpdateHybridBlah(friends, NUM_FRIENDS_VIS, observedVobs[i]);
    }
#endif // VISUAL_TRACK_FRIENDS
 
    if (doGPSOpponentUpdate) {
        // perform prediction update on each opponent structure
        for(int i = 0; i < NUM_OPPONENTS; i++)
            opponents[i]->growVariance(OPPONENT_VAR_GROWTH);

        for(int i = vobMinOpponent; i <= vobMaxOpponent; i++) {
            if(observedVobs[i].var < get95CF(150))
                applyKalmanUpdateHybrid(opponents, NUM_OPPONENTS, observedVobs[i]);
        }
    }
    
    motionForward = 0.0;
    motionLeft = 0.0;
    motionTurn = 0.0;
}

// This gets called by ResultCamera in Vision.cc. It uses the field line/edge
// methods from VisualCortex to do a measurement correction on the kalman
// filter.

void GPS::GPSEdgeUpdate(VisualCortex *vc, CommonSense *sensors) {
    MVec3 oldRobot, newRobot, grad;
    MVec2 projEdgeCentre;
    double newMatch;
    bool output;

    useLineData = false;

	if (isPenalized || !doGPSVisionUpdate)
		return;

    // Don't do line localisation if we don't have any points
    if (vc->numEdgePnts == 0) return;

    oldRobot(0, 0) = r[0].mean.pos.x;
    oldRobot(1, 0) = r[0].mean.pos.y;
    oldRobot(2, 0) = DEG2RAD(r[0].mean.h);

    output = false;
    DER_DEBUG(
        // This controls output of debugging stuff
        edgeDebugCounter++;

        if (edgeDebugCounter > 90) {
            edgeDebugCounter = 0;
            output = true;
        }
    );

    if (output) {
        cout << "==GPS::GPSEdgeUpdate==========================\n";
        //cout << "x\t\ty\t\th\t\tmatch\t\tlearn\n";
    }

    newMatch = gradientAscent(oldRobot, newRobot, vc);
        
    // Don't do the update when we get a perfect match with no movement,
    // since this is caused by only seeing field points (eg. when looking
    // down), which match to anywhere in the field area.
    if (newMatch == 1.0 && newRobot.equals(oldRobot)) return;

    vc->calcEdgeCentre(projEdgeCentre);
    
    int possibles, bestSymmetry; 
    double weights[MAX_EDGE_DUPLICITY], bestWeight;
    double posX[MAX_EDGE_DUPLICITY], varX[MAX_EDGE_DUPLICITY];
    double posY[MAX_EDGE_DUPLICITY], varY[MAX_EDGE_DUPLICITY];
    double posH[MAX_EDGE_DUPLICITY], varH[MAX_EDGE_DUPLICITY];
    MVec3 pos[MAX_EDGE_DUPLICITY], posVar[MAX_EDGE_DUPLICITY];

    possibles = calcAlternateLocations(newRobot, projEdgeCentre, pos, posVar);
    
    // If the head is moving quickly, then the match value probably won't be
    // as good (due to image distortion), so lower the minimum match
    // threshold. However, because of the distortion the final update accuracy
    // won't be as good either, so increase the udpate variance.
    double varScale = 1.0 + SQUARE(sensors->getHeadSpeed() * 50.0);
    double minMatch = 0.5 + 0.3 / (1.0 + sensors->getHeadSpeed() * 50);
        
    if (output) {
        cout << "Edge update with " << vc->numEdgePnts << " points.\n";
        cout << "Match " << newMatch << ", minimum match " << minMatch;
        cout << ", with " << possibles << " symmetries.\n";

        cout << "Original update at (";
        cout << newRobot(0, 0) << ", ";
        cout << newRobot(1, 0) << ", ";
        cout << RAD2DEG(newRobot(2, 0)) << ")\n";

        cout << "Edge centre at (";
        cout << projEdgeCentre(0, 0) << ", ";
        cout << projEdgeCentre(1, 0) << ")\n";
    }
        
    // Copy the vector array to individual component arrays
    int n = 0;
    bestSymmetry = 0; bestWeight = -1.0;
    for (int i = 0; i < possibles; i++) {
        weights[n] = vc->calcRobotMatch(pos[i], grad);
        
        // Punish updates that were off-field
        double testx = pos[i](0, 0);
        double testy = pos[i](1, 0);
        FieldClipping(testx, testy);
        double offFieldPenalty = SQUARE(testx - pos[i](0, 0))
                               + SQUARE(testy - pos[i](1, 0));
        weights[n] *= exp(-0.02 * offFieldPenalty);

        // Only use this observation if point match is reasonable
        if (weights[n] >= minMatch) {
            posX[n] = pos[i](0, 0); varX[n] = posVar[i](0, 0) * varScale;
            posY[n] = pos[i](1, 0); varY[n] = posVar[i](1, 0) * varScale;
            posH[n] = pos[i](2, 0); varH[n] = posVar[i](2, 0) * varScale;
            posH[n] = RAD2DEG(posH[n]);
            varH[n] = RAD2DEG(RAD2DEG(varH[n]));
        
            if (weights[n] > bestWeight) {
                bestSymmetry = i;
                bestWeight = weights[n];
            }
        
            if (output) {
                cout << "Update " << n << ": mean (";
                cout << posX[n] << ", " << posY[n] << ", " << posH[n];
                cout << ")    variance (";
                cout << varX[n] << ", " << varY[n] << ", " << varH[n];
                cout << ")    weight " << weights[n] << "\n";
            }
            
            n++;
        }
    }

    // Apply the edge updates to our self position.
    if (n > 0) {
        GPSDirectPositionUpdate(n, posX, posY, posH, varX, varY, varH, weights);
        useLineData = true;
    
        // This sets the world model display to show the edge points positions
        // with respect to the best symmetry
        vc->calcRobotMatch(pos[bestSymmetry], grad);
        directx = pos[bestSymmetry](0, 0);
        directy = pos[bestSymmetry](1, 0);
        directh = RAD2DEG(pos[bestSymmetry](2, 0));

    } else {
        vc->calcRobotMatch(newRobot, grad);
        directx = -1;
        directx = -1;
        directx = -1;
    }
}

// A very expensive operation (will cause dropped frames), this was put in for
// the localisation challenge player. Basically performs a gradient ascent for
// many points on the field. And sets the gaussian mode array to the top
// matching end locations. This should only really be used if there is a
// decent amount of edge point data available, eg. if stationary mapping has
// been running for a while.

void GPS::GPSGlobalMaxEdgeSet(VisualCortex *vc) {
    MVec3 startPos, endPos;
    MMatrix3 newVar(500.0);
    SingleGaussian temp;
    
    newVar(2, 2) = 300.0;
    numGaussians = 0;
    
    for (int x = 1; x < 4; x++) {
        for (int y = 1; y < 4; y++) {
            for (int h = 0; h < 8; h++) {
                startPos(0, 0) = x * (FIELD_WIDTH / 4.0);
                startPos(1, 0) = y * (FIELD_LENGTH / 4.0);
                startPos(2, 0) = DEG2RAD(h * 45.0);
                
                temp.weight = gradientAscent(startPos, endPos, vc);
                temp.mean.pos.setVector(vCART, endPos(0, 0), endPos(1, 0));
                temp.mean.h = NormalizeAngle_0_360(RAD2DEG(endPos(2, 0)));
                temp.cov = newVar;

                insertGaussian(temp);
            }   
        }
    }

    formDistribution();
    GPSOutputSelfData();    
}

// Perform a gradient ascent to highest match value

static const int maxAscentSteps = 100;

double GPS::gradientAscent(MVec3 &originalRobot, MVec3 &newRobot, VisualCortex *vc) {
    MVec3 oldRobot, oldGradient, newGradient, update;
    double oldMatch, newMatch, angleScale, learnRate;
    bool output;

    angleScale = 0.0005;
    learnRate = 2000.0;
    
    oldMatch = 0.0;
    oldRobot = originalRobot;
    newMatch = oldMatch;
    newRobot = oldRobot;
    newGradient(0, 0) = 0.0;
    newGradient(1, 0) = 0.0;
    newGradient(2, 0) = 0.0;
    for (int i = 0; i < maxAscentSteps && learnRate > 1.0; i++) {    
        // Calculate match and match gradient at current point
        newMatch = vc->calcRobotMatch(newRobot, newGradient);
        
        //if (output) {
        //    cout << newRobot(0, 0) << "\t" << newRobot(1, 0) << "\t" << newRobot(2, 0);
        //    cout << "\t" << (newMatch / totalWeight) << "\t" << learnRate << "\n";
        //}
        
        if (newMatch < oldMatch) {
            // If the new match value is worse than the old match value, then we
            // must have overshot a peak. So don't use the new position &
            // decrease the learn rate.
            learnRate *= 0.5;

            newMatch = oldMatch;
            newRobot = oldRobot;
            newGradient = oldGradient;
        } else {
            // Alternatively, if the match is better, we may be moving too
            // slowly, and so gradually accelerate by increasing learn rate.
            learnRate *= 1.5;
            
            oldMatch = newMatch;
            oldRobot = newRobot;
            oldGradient = newGradient;
        }

        // Move in gradient direction
        update = newGradient;
        update(2, 0) *= angleScale;
        update *= learnRate;

        newRobot += update;
    }
    
    return oldMatch;
}

// Derrick: The following commented code is an alternate form of field line
// localisation I was developing until I fixed a serious bug in the old line
// localisation that made it work much better. Currently it doesn't seem to
// work, but I included it here for future reference.

/*
// Performs correction update according to field line matching. For each
// gaussian in the distribution, it generates a set of 27 sample points, based
// on that gaussian's mean and variance. At each sample point, the likelyhood
// of the field lines/edges seen matching the position is calculated.
// A new gaussian(s) is then generated from the sample points, and applied as a
// direct kalman update.

void GPS::GPSEdgeUpdate(VisualCortex *vc, CommonSense *sensors) {
    int i, oldNumGaussians, mode, obsMode, numObs, dimCount;
	double jacX[3], jacY[3], jacH[3], upVar[3], iv[3];
    double oldWeight, baseVar, infoVal;
    MVec3 samplePoint, grad;
    SingleGaussian temp;

	if (isPenalized || !doGPSVisionUpdate)
		return;
    
    output = false;
    if (vc->numEdgePnts < 1) return;

    infoVal = (1.0 - exp(-0.005 * SQUARE(vc->numEdgePnts)));
    infoVal /= (1.0 + 20.0 * SQUARE(sensors->getHeadSpeed()));
    baseVar = 1.0 / infoVal;

    if (infoVal < 0.5) return;

    DER_DEBUG(
        edgeDebugCounter++;
        if (edgeDebugCounter > 50) {
            output = true;
            edgeDebugCounter = 0;
        }
    );
    
    // Work from a backup of the old distribution
    for (i = 0; i < numGaussians; i++) {
        rCopy[i] = r[i];
        r[i].weight *= noiseThreshold;
    }
    
    oldNumGaussians = numGaussians;

    if (output) {
        cout << "GPS::GPSEdgeUpdate localising with " << vc->numEdgePnts;
        cout << " points, " << oldNumGaussians << " gaussians.\n";
        cout << "Information value is: " << infoVal << "\n";
        cout << "Minimum variance is: " << baseVar << "\n";
    }

    for (mode = 0; mode < oldNumGaussians; mode++) {
        oldWeight = rCopy[mode].weight;

        // Calculate these convenience values
        curMean(0, 0) = rCopy[mode].mean.pos.x;
        curMean(1, 0) = rCopy[mode].mean.pos.y;
        curMean(2, 0) = DEG2RAD(rCopy[mode].mean.h);
        curDev(0, 0) = sqrt(rCopy[mode].cov(0, 0));
        curDev(1, 0) = sqrt(rCopy[mode].cov(1, 1));
        curDev(2, 0) = DEG2RAD(sqrt(rCopy[mode].cov(2, 2)));
        
        // Sample the current gaussian distribution
        if (output) cout << "Sample matrix is:\n";
        for (i = 0; i < 27; i++) {
            calcSamplePoint(i, samplePoint);                        
            samples[i] = vc->calcRobotMatch(samplePoint, grad);
            if (output) {
                cout << samples[i] << "  ";
                if ((i % 9) == 8) cout << "\n";
            }            
        }
        
        // Create a set of observation gaussians, based on the sampling
        numObs = reformGaussian();

        if (output) {
            cout << "Gaussian " << mode << " reformed into ";
            cout << numObs << " observation(s)\n";
        }
        
        // Kalman update the current gaussian with every observation mode.
        for (obsMode = 0; obsMode < numObs; obsMode++) {
            dimCount = 0;

            vc->calcRobotMatch(mean[obsMode], grad);
            
            if (output) {
                cout << "Observation " << obsMode << " has prob ";
                cout << prob[obsMode] << "\n";
                cout << "Mean:\n";
                mean[obsMode].printOut();
                cout << "Variance:\n";
                var[obsMode].printOut();
            }
        
            // Create the jacobian and innovation vector for the update
            for (i = 0; i < 3; i++) {
                if(var[obsMode](i, 0) < LARGE_VAL / 2) {
                    upVar[dimCount] = var[obsMode](i, 0);
                    iv[dimCount] = mean[obsMode](i, 0) - curMean(i, 0);
                    
                    jacX[dimCount] = 0;
                    jacY[dimCount] = 0;
                    jacH[dimCount] = 0;
                    
                    if (i == 0)
                        jacX[dimCount] = 1;
                    else if (i == 1)
                        jacY[dimCount] = 1;
                    else if (i == 2) {
                        jacH[dimCount] = 1;
                                        
                        // If this dimension is the heading, have to convert back
                        // to degrees and normalise innovation vector
                        iv[dimCount] = RAD2DEG(iv[dimCount]);
                        iv[dimCount] = NormalizeAngle_180(iv[dimCount]);
                        
                        // Need to do RAD2DEG twice because upVar is a variance
                        upVar[dimCount] = RAD2DEG(RAD2DEG(upVar[dimCount]));
                    }

                    upVar[dimCount] += baseVar;
        
                    dimCount++;
                }
            }
            
            temp = rCopy[mode];
            
            if (prob[obsMode] > 0.6) {
                switch(dimCount) {
                case 1: kalmanUpdate<1>(&temp, jacX, jacY, jacH, iv, upVar); break;
                case 2: kalmanUpdate<2>(&temp, jacX, jacY, jacH, iv, upVar); break;
                case 3: kalmanUpdate<3>(&temp, jacX, jacY, jacH, iv, upVar); break;
                }
            }
                        
            // Use peak match value instead of kalman weight scaler to
            // determine new weight
            temp.weight = oldWeight * prob[obsMode];
            
            // Insert the corrected mode into the new distribution
            insertGaussian(temp);
        }
    }
    
    formDistribution();
}

// From the sampling of the current gaussian, this creates a set of observation
// gaussians. There is one observation for every local maximum in the sampling
// array. The number of observation gaussians is returned. The mean of each
// observation is put in the x, y, h arrays, variances are put in the xV, yV, hV
// arrays, and the weight scaler for each observation is put in the prob array.

int GPS::reformGaussian() {
    int maxIndices[NUM_SAMPLES], i, j, numObs, numMax;
    int numNeighbours, neighbours[2];
    double logRatio, peakRatio, sideRatio;

    // Find all local maximums in the sampling array
    numMax = findSampleMaximums(maxIndices);
    
    // For each maximum, calculate a mean and variance
    for (i = 0; i < numMax; i++) {
        if (output) cout << "Local max at index " << maxIndices[i] << "\n";

        calcSamplePoint(maxIndices[i], mean[i]);
        prob[i] = samples[maxIndices[i]];

        for (j = 0; j < 3; j++) {
            numNeighbours = calcNeighbourIndices(j, maxIndices[i], neighbours);
            // If this local maximum only has one neighbour along the current
            // dimension, then assume that maximum match occurs (along this
            // dimension) at the local max.
            if (numNeighbours == 1) neighbours[1] = neighbours[0];
                
            // Note: this can make the dog crash if there are sample values
            // of zero, so make sure there aren't! A likely canditate for
            // getting zero samples is the INFINITY_GRADIENT_VALUE constant
            // in VisualCortex: if it's zero, try setting it to a small value,
            // eg. 1E-8
            sideRatio = prob[i] / samples[neighbours[1]];
            peakRatio = prob[i] * prob[i];
            peakRatio /= (samples[neighbours[0]] * samples[neighbours[1]]);

            if (output) {
                cout << "Dir " << j << ": peakRatio is " << peakRatio;
                cout << "  sideRatio is " << sideRatio << "\n";
            }

            logRatio = log(peakRatio);
            if (logRatio != 0.0) {
                // How did I get these formulas? If you have three function
                // values f(x1), f(x2), f(x3), then you can fit a gaussian
                // distribution to them as long as x1, x2, x3 are equally
                // spaced. So assume the function values lie on a gaussian with
                // unknown mean and variance and solve simultaneously.
                logRatio = 1.0 / logRatio;
                var[i](j, 0) = curDev(j, 0) * curDev(j, 0) * logRatio;
                logRatio *= log(sideRatio);
                mean[i](j, 0) += curDev(j, 0) * (0.5 - logRatio);

            } else var[i](j, 0) = LARGE_VAL;
        }
    }
    
    // Sometimes there will be two local maximums at same x, y position but one
    // with -1 standard deviation heading and the other at +1 s.d. heading.
    // Would rather merge these two into a centrally located maximum.
    int hInd, xyInd, otherhInd, otherxyInd;
    double totalProb;

    numObs = 0;
    for (i = 0; i < numMax; i++) {
        hInd = maxIndices[i] % 3;
        xyInd = maxIndices[i] / 3;
        otherhInd = maxIndices[i + 1] % 3;
        otherxyInd = maxIndices[i + 1] / 3;
        
        if (i < numMax - 1) {
            // If a +-heading pair exists, average them
            if (xyInd == otherxyInd && hInd == 0 && otherhInd == 2) {
                totalProb = prob[i] + prob[i + 1];
                mean[i] *= prob[i]; mean[i + 1] *= prob[i + 1];
                var[i]  *= prob[i]; var[i + 1]  *= prob[i + 1];

                mean[i + 1] += mean[i];
                var[i + 1] += var[i];
                
                mean[i + 1] /= totalProb;
                var[i + 1] /= totalProb;
                prob[i + 1] = totalProb / 2;                
        
                i++;
            }
        }
        
        // If we have merged any observations, have to move the other
        // observations up in the array
        if (i > numObs) {
            prob[numObs] = prob[i];
            mean[numObs] = mean[i];
            var[numObs] = var[i];
        }
        
        numObs++;
    }
    
    return numObs;
}

// Returns the sample indices of all local maximums in the sample array

int GPS::findSampleMaximums(int *indices) {
    int i, j, k, numNeighbours, neighbours[2], numMax;
    bool isMax;
    
    numMax = 0;
    for (i = 0; i < NUM_SAMPLES; i++) {
        isMax = true;
    
        // To be a local maximum, the sample value has to be greater than the
        // values of its neighbours. Check neighbours in all three directions.
        for (j = 0; j < 3; j++) {
            numNeighbours = calcNeighbourIndices(j, i, neighbours);
            
            if (numNeighbours > 1) {
                for (k = 0; k < numNeighbours; k++)
                    isMax = isMax && (samples[i] >= samples[neighbours[k]]);
                
            // If neighbouring sample has the same match value, only the sample
            // closer to the centre will be accepted as a local maximum
            } else isMax = isMax && (samples[i] > samples[neighbours[0]]);
        }
        
        if (isMax)
            indices[numMax++] = i;
    }

    return numMax;
}

// Returns the robot coordinates of the given sample index

void GPS::calcSamplePoint(int index, MVec3 &out) {
    int temp, i;

    out = curMean;
    temp = index;
    
    // Calculate sample points to be 1 standard deviation away from mean,
    // aligned along the x, y, h axes.
    for (i = 2; i >= 0; i--) {
        out(i, 0) += curDev(i, 0) * (double)((temp % 3) - 1);
        temp /= 3;
    }
}

// Returns the number of neighbours a sample has in the specified direction.
// dir = 0 for x axis, 1 for y, 2 for h. Neighbour indices are also returned.

int GPS::calcNeighbourIndices(int dir, int index, int *outIndices) {
    int i, temp, d[3], dInd, neighbours;
    
    // Calculate individual coordinate indices from sample index
    temp = index;
    for (i = 2; i >= 0; i--) {
        d[i] = (temp % 3) - 1;
        temp /= 3;
    }
    
    // Calculate the offset from the sample index the neighbours should be at
    dInd = 1;
    if (dir == 0) dInd = 9;
    if (dir == 1) dInd = 3;
    
    // Check if neighbours are still inside the sample array
    neighbours = 0;
    if (d[dir] > -1) outIndices[neighbours++] = index - dInd;
    if (d[dir] < 1) outIndices[neighbours++] = index + dInd;
    
    return neighbours;
}
*/

void GPS::beaconDistanceCal() {

}


//Copies some things from the array of VisualObjects we are given
//into our own data structure.
//This is done so that if VisualObjects change, it is easier to change
//gps code.
void GPS::UpdateObservedVobs(VisualObject * nvo) {
    for (int i = 0; i < VOB_COUNT; i++) {
        observedVobs[i].vobType = i;
        observedVobs[i].cf = nvo[i].cf;
        observedVobs[i].var = nvo[i].var;
        if(observedVobs[i].cf > 0) {
            observedVobs[i].h   = nvo[i].head;
            observedVobs[i].d   = nvo[i].getFlatDistance();
            observedVobs[i].angleVar = nvo[i].angleVar;
        } else {
            observedVobs[i].h   = 0;
            observedVobs[i].d   = 0;
            observedVobs[i].angleVar = LARGE_VAL;
			observedVobs[i].var = LARGE_VAL;
#ifdef VALGRIND
			VALGRIND_MAKE_WRITABLE(&observedVobs[i].h,sizeof(double));
			VALGRIND_MAKE_WRITABLE(&observedVobs[i].d,sizeof(double));
			VALGRIND_MAKE_WRITABLE(&observedVobs[i].angleVar,sizeof(double));
			// note - var could be used
#endif // VALGRIND
        }
    }
}

// Adds another mode to the distribution, making sure it is inserted into the
// correct (weight sorted) position.

void GPS::insertGaussian(SingleGaussian &temp) {
    int i;
    
    for (i = numGaussians; i > 0 && temp.weight > r[i - 1].weight; i--) {
        if (i < maxGaussians) r[i] = r[i - 1];
    }
            
    if (i < maxGaussians) {
        r[i] = temp;
        r[i].mean.updateVar(&r[i].cov);

        if (numGaussians < maxGaussians)
            numGaussians++;
    }
}

// Normalises the sum of gaussians, culls and merges individual gaussians,
//  also sorts gaussians by decreasing weight

void GPS::formDistribution() {
    int i, j, k, l, cullCount, mergeCount, oldGaussians;
    double cullWeight;
    double dist, deltaX, deltaY, deltaH;
    SingleGaussian temp;
    MMatrix3 deltaCov;
    
    oldGaussians = numGaussians;
    cullCount = 0; mergeCount = 0;
    
    // Merge gaussians that have similar means and covariances
    for (i = 1; i < numGaussians; i++) {
        if (r[i].weight != 0.0) {

            for (j = 0; j < i; j++) {
                if (r[j].weight != 0.0) {
                    deltaX = r[i].mean.pos.x - r[j].mean.pos.x;
                    deltaY = r[i].mean.pos.y - r[j].mean.pos.y;
                    deltaH = NormalizeAngle_180(r[i].mean.h - r[j].mean.h);
                    dist = deltaX * deltaX + deltaY * deltaY + deltaH * deltaH;
                    
                    if (dist < mergeThreshold) {
                        deltaCov = r[i].cov; 
                        deltaCov -= r[j].cov;
                        
                        dist = 0.0;
                        for (k = 0; k < 3; k++) {
                            for (l = 0; l < 3; l++)
                                dist += deltaCov(k, l) * deltaCov(k, l);
                        }
                        
                        if (dist < mergeCovThreshold) {
                            mergeGaussians(i, j);
                            mergeCount++;
                        }
                    }
                }
            }
        }
    }

    // Insertion sort. Since array is probably close to being sorted, this
    // should be quite efficient.
    for (i = 1; i < numGaussians; i++) {
        if (r[i - 1].weight < r[i].weight) {
            temp = r[i];
            
            for (j = i - 1; j >= 0 && r[j].weight < temp.weight; j--)
                r[j + 1] = r[j];
                
            r[j + 1] = temp;
        }
    }

    normaliseDistribution();
     
    // Cull unlikely gaussians and renormalise again
    /*if (!useRelativeCullWeight) {
        cullWeight = cullThreshold;
    } else {
        double maxWeight = r[0].weight;
        double minWeight = r[0].weight;
        for (i = 0; i < numGaussians; i++) {
            if (r[i].weight > maxWeight) maxWeight = r[i].weight;
            if (r[i].weight < minWeight) minWeight = r[i].weight;
        }

        cullWeight = (maxWeight + minWeight) * 0.5;
        if (cullWeight > cullThreshold) cullWeight = cullThreshold;
    }
		
    for (i = 0; i < numGaussians; i++) {
        if (r[i].weight < cullWeight) {
            r[i].weight = 0.0;
            cullCount++;
        }
    }

    normaliseDistribution();*/

    // Remove zero and NaN weight modes from the mode count
    while (numGaussians > 0 && (r[numGaussians - 1].weight == 0.0 ||
           r[numGaussians - 1].weight != r[numGaussians - 1].weight))
    	numGaussians--;

    // If all modes have been removed, reset distribution. If this is being
    // called, then something is wrong.
    if (numGaussians == 0) {
        cout << "GPS::formDistribution warning: Total probability zero, resetting.\n";
        
        numGaussians = 1;
        r[0].mean.pos.setVector(vCART, FIELD_WIDTH / 2, FIELD_LENGTH / 2);
        r[0].mean.h = HALF_CIRCLE/2;
        r[0].cov(0,0)= get95CF (FIELD_WIDTH / 2);
        r[0].cov(0,1)= 0.0;
        r[0].cov(0,2)= 0.0;
        r[0].cov(1,0)= 0.0;
        r[0].cov(1,1)= get95CF (FIELD_LENGTH / 2);
        r[0].cov(1,2)= 0.0;
        r[0].cov(2,0)= 0.0;
        r[0].cov(2,1)= 0.0;
        r[0].cov(2,2)= get95CF (HALF_CIRCLE);
	    r[0].weight = 1.0;
	    r[0].mean.updateVar(&r[0].cov);
    }
}

// Makes sure the total probability of the sum of gaussians is 1

void GPS::normaliseDistribution() {
    double totalWeight = 0.0;
    int i;

    for (i = 0; i < numGaussians; i++) 
        totalWeight += r[i].weight;
    
    if (totalWeight != 0.0) {    
        for (i = 0; i < numGaussians; i++)
            r[i].weight /= totalWeight;
    }
}

// Merges two gaussians into one

inline void GPS::mergeGaussians(int a, int b) {
    double totalWeight, scaleA, scaleB;
    
    totalWeight = r[a].weight + r[b].weight;
    if (totalWeight == 0.0) return;
    
    scaleA = r[a].weight / totalWeight;
    scaleB = r[b].weight / totalWeight;
    
    r[a].mean.pos.scale(scaleA); r[a].mean.h *= scaleA; r[a].cov *= scaleA;
    r[b].mean.pos.scale(scaleB); r[b].mean.h *= scaleB; r[b].cov *= scaleB;
    
    r[a].mean.pos.add(r[b].mean.pos);
    r[a].mean.h += r[b].mean.h;
    r[a].cov += r[b].cov;
    r[a].weight += r[b].weight;
    r[a].mean.updateVar(&r[a].cov);
    
    r[b].weight = 0.0;   
}

// Derrick: For debugging.
// Does a small update test

void GPS::testKalman() {
    SetGoals(true);

	// Set robot position to be two equally probable gaussians
    numGaussians = 2;
    r[0].mean.pos.setVector(vCART, 50.0, 150.0);
    r[0].mean.h = 0;
    r[0].cov(0,0)= 100;
    r[0].cov(0,1)= 0;
    r[0].cov(0,2)= 0;
    r[0].cov(1,0)= 0;
    r[0].cov(1,1)= 100;
    r[0].cov(1,2)= 0;
    r[0].cov(2,0)= 0;
    r[0].cov(2,1)= 0;
    r[0].cov(2,2)= 100;
	r[0].weight = 1.0;
    r[0].mean.updateVar(&r[0].cov);
    r[1].mean.pos.setVector(vCART, 115.0, 115.0);
    r[1].mean.h = 315;
    r[1].cov(0,0)= 100;
    r[1].cov(0,1)= 0;
    r[1].cov(0,2)= 0;
    r[1].cov(1,0)= 0;
    r[1].cov(1,1)= 100;
    r[1].cov(1,2)= 0;
    r[1].cov(2,0)= 0;
    r[1].cov(2,1)= 0;
    r[1].cov(2,2)= 100;
	r[1].weight = 1.0;
    r[1].mean.updateVar(&r[1].cov);
    
    cout << "--- Start GPS Weight function testing\n";
    formDistribution();
    
    /*// Set observation heading and distance
    SlimVisOb vi;
    vi.vobType = 8;
    vi.d = 220.0;
    vi.var = 100.0;
    vi.h = 0.0;
    vi.angleVar = 100.0;
    vi.cf = 1.0;
   
    double iv[2], jacX[2], jacY[2], jacH[2], var[2];
    
    // Do an update based on the observation
    for (int mode = 0; mode < 2; mode++) {
        //cout <<"For Gaussian no." << mode << "\n";
        for (int i = 0; i < 2; i++) {
            var[i] = 0.0;
            iv[i] = 0.0;
            jacX[i] = 0.0;
            jacY[i] = 0.0;
            jacH[i] = 0.0;
        }
        
        cout << "Adding measurement...\n";
        addMeasurement(&r[mode], vi, jacX, jacY, jacH, iv, var, 0, true);
        cout << "Updating kalmann filter...\n";
        kalmanUpdate<2>(&r[mode], jacX, jacY, jacH, iv, var);
        cout << "Updating variance...\n";
        r[mode].mean.updateVar(&r[mode].cov);    
    }*/
    
    // Set multiple observation locations
    double x[4], y[4], h[4], vX[4], vY[4], vH[4], w[4];
    x[0] = 80 ; y[0] = 140; h[0] = 10 ; vX[0] = 100; vY[0] = 250; vH[0] = 90 ;
    x[1] = 180; y[1] = 40 ; h[1] = 310; vX[1] = 180; vY[1] = 200; vH[1] = 110;
    x[2] = 50 ; y[2] = 60 ; h[2] = 10 ; vX[2] = 200; vY[2] = 200; vH[2] = 150;
    x[3] = 180; y[3] = 40 ; h[3] = 310; vX[3] = 180; vY[3] = 200; vH[3] = 110;
    w[0] = 1; w[1] = 1; w[2] = 1; w[3] = 1;
    GPSDirectPositionUpdate(4, x, y, h, vX, vY, vH, w);
     
    formDistribution();
    cout << "--- End GPS Weight function testing\n";   

	// Robot position probability starts with only one gaussian
    numGaussians = 1;
    r[0].mean.pos.setVector(vCART, FIELD_WIDTH/2, FIELD_LENGTH/2);
    r[0].mean.h = HALF_CIRCLE/2;
    r[0].cov(0,0)= get95CF(FIELD_WIDTH);
    r[0].cov(1,1)= get95CF(FIELD_LENGTH);
    r[0].cov(2,2)= get95CF(HALF_CIRCLE);
	r[0].weight = 1.0;
	r[0].mean.updateVar(&r[0].cov);
}

void GPS::addMeasurement(SingleGaussian *g, SlimVisOb &vi,
                double *jacWRTX, double *jacWRTY, double *jacWRTH,
                double *iv, double *vars, int curNum, bool dist) {

	// coordinates of object
    double bx = (*coord)[vi.vobType - 1][0];
    double by = (*coord)[vi.vobType - 1][1];
                
    addMeasurement(g, vi, bx, by, jacWRTX, jacWRTY, jacWRTH,
                    iv, vars, curNum, dist);
}

inline void GPS::addMeasurement(SingleGaussian *g, SlimVisOb &vi, double bx, double by,
                double *jacWRTX, double *jacWRTY, double *jacWRTH,
                double *iv, double *vars, int curNum, bool dist) {
		
	double dx = g->mean.pos.x - bx;
    double dy = g->mean.pos.y - by;
    double dxsq = SQUARE(dx);
    double dysq = SQUARE(dy);
	
	double expHead = RAD2DEG(atan2(-dy,-dx)) - g->mean.h;
    double measHead = vi.h;
	double expDist = sqrt(dxsq + dysq);

    // keep expected heading and measurement within 180 of each other
    ANGLE_FIX(expHead);
    if(expHead - measHead > HALF_CIRCLE)
        measHead += FULL_CIRCLE;
    else if(measHead - expHead > HALF_CIRCLE)
        measHead -= FULL_CIRCLE;
	
	// add to innovation vector
	iv[curNum] = measHead - expHead;
	if(dist) iv[curNum+1] = vi.d - expDist;
	
	// add to jacobians
	jacWRTX[curNum] = -(DEG_OVER_RAD*dy)/(dxsq + dysq);
	jacWRTY[curNum] = (DEG_OVER_RAD*dx)/(dxsq + dysq);
	jacWRTH[curNum] = -1;
	if(dist) {
		jacWRTX[curNum+1] = dx/expDist;
		jacWRTY[curNum+1] = dy/expDist;
		jacWRTH[curNum+1] = 0;
	}
	
	// add to variances
	vars[curNum] = vi.angleVar;
	if(dist) vars[curNum+1] = 8*vi.var;
				
}


template <int n> void 
GPS::kalmanUpdate(SingleGaussian *g, double *jacWRTX, double *jacWRTY, double *jacWRTH, double *iv, double *vars) {

	// jacobian of measurement
	MMatrix<double,n,3> jac(0);
	for(int i=0 ; i<n ; i++) {
		jac(i,0) = jacWRTX[i];
		jac(i,1) = jacWRTY[i];
		jac(i,2) = jacWRTH[i];
	}
	MMatrix<double,3,n> jacT;
	jacT.isTranspose(jac);
	
	// innovation vector of measurement
	MMatrix<double,n,1> innVec(0);
	for(int i=0 ; i<n ; i++)
		innVec(i,0) = iv[i];
	   	
	// compute kalman gain
	MMatrix<double,n,3> jCov;
	jCov.isMult(jac, g->cov);
	MMatrix<double,n,n> obsCov;
	obsCov.isMult(jCov,jacT);
	for(int i=0 ; i<n ; i++)
		obsCov(i,i) += vars[i];
	MMatrix<double,3,n> covjT;
	covjT.isMult(g->cov,jacT);
	MMatrix<double,3,n> k;
	MMatrix<double,n,n> obsCovInv;
	obsCovInv.isInverse(obsCov);
	k.isMult(covjT,obsCovInv);
    
	// Calculate new state
	MVec3 priori;
    priori(0, 0) = g->mean.pos.x;
	priori(1, 0) = g->mean.pos.y;
	priori(2, 0) = g->mean.h;
	MVec3 move;
	move.isMult(k,innVec);
	
    // Calculate new weight
    double exponent, weightScaler;
    MMatrix<double,n,1> exponentVec(innVec);
    exponentVec.preMult(obsCovInv);

    exponent = 0.0;
    for (int i = 0; i < n; i++)
        exponent += innVec(i, 0) * exponentVec(i, 0);
    exponent *= -0.5;

    /*double determinant = obsCov.Det();
    if (determinant == 0.0) weightScaler = 0.0;
    else weightScaler = exp(exponent) / sqrt(fabs(determinant));*/
    weightScaler = exp(exponent);
    
    // Calculate new covariances
    MMatrix3 m(*unitMatrix3);
    MMatrix3 kJ;
    kJ.isMult(k,jac);
    m -= kJ;
    g->cov.preMult(m);

    MVec3 newState(priori);
    newState += move;
    double rx = newState(0, 0);
    double ry = newState(1, 0);
    FieldClipping(rx, ry);
    g->mean.pos.setVector(vCART, rx, ry);
    g->mean.h = newState(2, 0);
    ANGLE_FIX(g->mean.h);

    // If we got a bad measurement, send a cplane
    if (weightScaler < noiseThreshold)
        sendCPlane = true;

    g->weight *= weightScaler;

    nX[n]++;
    sumX[n] += weightScaler;
    sumXSquared[n] += SQUARE(weightScaler);
    
    // Visual pull has been disabled by Derrick Apr 2004
    visPull(0, 0) = 0; visPull(1, 0) = 0; visPull(2, 0) = 0;
    vbPull(0, 0) = 0; vbPull(1, 0) = 0;
}

void GPS::GPSDirectPositionUpdate(double x, double y, double h,
                double xvar, double yvar, double hvar) {
    double unitWeight = 1.0;
    GPSDirectPositionUpdate(1, &x, &y, &h, &xvar, &yvar, &hvar, &unitWeight);
}

void GPS::GPSDirectPositionUpdate(int numObs, double *x, double *y, double *h,
                double *xvar, double *yvar, double *hvar, double *weight) {

	if (isPenalized || !doGPSVisionUpdate)
		return;
     
    directx = x[0];
    directy = y[0];
    directh = h[0];
	
    SingleGaussian temp;
	
	double jacX[3], jacY[3], jacH[3];
	double var[3], iv[3];
    double minWeight;

    int i, oldNumGaussians, mode, obsMode, currentDimCount = 0;
    
    if (numObs < 1) return;
    
    // Work from a backup of the old distribution
    minWeight = r[numGaussians - 1].weight;
    for (i = 0; i < numGaussians; i++) {
        rCopy[i] = r[i];
        r[i].weight *= noiseThreshold;
    }
    
    oldNumGaussians = numGaussians;
        	
    for (obsMode = 0; obsMode < numObs; obsMode++) {
        currentDimCount = 0;
        
        if(xvar[obsMode] < LARGE_VAL/2) {
            jacX[currentDimCount] = 1;
            jacY[currentDimCount] = 0;
            jacH[currentDimCount] = 0;
            var[currentDimCount] = xvar[obsMode];
            currentDimCount++;
        }
            
        if(yvar[obsMode] < LARGE_VAL/2) {
            jacX[currentDimCount] = 0;
            jacY[currentDimCount] = 1;
            jacH[currentDimCount] = 0;
            var[currentDimCount] = yvar[obsMode];
            currentDimCount++;                
        }
            
        if(hvar[obsMode] < LARGE_VAL/2) {
            jacX[currentDimCount] = 0;
            jacY[currentDimCount] = 0;
            jacH[currentDimCount] = 1;
            var[currentDimCount] = hvar[obsMode];
            currentDimCount++;
        }
   
        mode = 0;    
        while (mode < oldNumGaussians && rCopy[mode].weight >= minWeight) {	
            // Calculate the effect of every observation mode on each of
            // the distribution modes
       
            currentDimCount = 0;

            if(xvar[obsMode] < LARGE_VAL/2)
                iv[currentDimCount++] = x[obsMode] - rCopy[mode].mean.pos.x;
            if(yvar[obsMode] < LARGE_VAL/2)
                iv[currentDimCount++] = y[obsMode] - rCopy[mode].mean.pos.y;
            if(hvar[obsMode] < LARGE_VAL/2)
                iv[currentDimCount++] =
                    NormalizeAngle_180(h[obsMode] - rCopy[mode].mean.h);
            
            temp = rCopy[mode];
            temp.weight *= weight[obsMode];
            
            switch(currentDimCount) {
            case 1: kalmanUpdate<1>(&temp, jacX, jacY, jacH, iv, var); break;
            case 2: kalmanUpdate<2>(&temp, jacX, jacY, jacH, iv, var); break;
            case 3: kalmanUpdate<3>(&temp, jacX, jacY, jacH, iv, var); break;
            }
            
            // Insert the corrected mode into the new distribution
            insertGaussian(temp);
            if (numGaussians == maxGaussians)
                minWeight = r[numGaussians - 1].weight;
            
            mode++;
        }
    }
    
    formDistribution();
}




int VisionUpdateCount = 0; //for debugging only



const WMObj& GPS::self() {
   return r[0].mean;
}

const MMatrix3& GPS::selfCov() {
	return r[0].cov;
}

unsigned int& GPS::sharedBehaviour() {
	return r[0].mean.behavioursVal;
}

const MVec3& GPS::getVisualPull() {
	return visPull;
}

const Vector& GPS::getBall(int context) {

	return newball->getPos(context);

}

const Vector& GPS::getVBall(int context) {
	
	return newball->getVelocity(context);
	
}

double GPS::getBallAngle() {

	const Vector& ball = newball->getPos(LOCAL);
	const Vector& vball = newball->getVelocity(LOCAL);
	
	double dp = ball.x*vball.x + ball.y*vball.y;
	double denom = ball.d * vball.d;
	
	if(abs(dp) > denom) {
		cout << "Error: length of vectors multplied together less than dot product" << endl;
		if(dp > 0)
			return 0;
		else
			return 1;
	}
	
	// hopefully won't happen
	if(abs(denom) < 0.0001) {
		return 180;
	}
	
	return RAD2DEG(acos(dp/denom));
			
}

const MMatrix2& GPS::getBallCov() {

	return newball->getCov();

}

double GPS::getBallMaxVar() {

	const MMatrix2& bp = newball->getCov();
	double b = -(bp(0,0) + bp(1,1));
    double c = bp(0,0)*bp(1,1) - bp(0,1)*bp(1,0);
    double posVar;
    if(b*b-4*c < 1E-5) {
        posVar = (-b+0)/2;
        if (b*b-4*c < -1E-5)
            cout << "WARNING: b*b-4*c < 0 (location 1): " << (b*b-4*c) << endl;
    } else {
        posVar = (-b + sqrt(b*b-4*c)) / 2;
    }
    return posVar;
}

const MMatrix2& GPS::getVelCov() {

	return newball->getVelCov();

}

double GPS::getVBallMaxVar() {

	const MMatrix2& bp = newball->getVelCov();
	double b = -(bp(0,0) + bp(1,1));
    double c = bp(0,0)*bp(1,1) - bp(0,1)*bp(1,0);
    double posVar;
    if(b*b-4*c < 1E-5) {
        posVar = (-b+0)/2;
        if (b*b-4*c < -1E-5)
            cout << "WARNING: b*b-4*c < 0 (location 1): " << (b*b-4*c) << endl;
    } else {
        posVar = (-b + sqrt(b*b-4*c)) / 2;
    }
    return posVar;
}

double GPS::getVBIVLength() {

	return newball->getIVLength();

}

MMatrix2 GPS::getVCLocal() {

	return newball->getVCovLocal();

}

MMatrix2 GPS::getVCBall() {

	return newball->getVCovBall();

}

// This function returns the teammate information. What is cache?
// If in the current frame, this function is called several times,
// instead of computing the result again simply return the cache result.
//
WMObj& GPS::teammate(int botNum, char origin) {

    // Check cache.
    if (teammateAccessId==frameId && botNum==lastTeammateAccessed && origin == lastTeammateOrigin) {
		//cout << "using cached teammate for teammate " << botNum << endl;
		return retTeammate;
	}
	teammateAccessId = frameId;
	lastTeammateAccessed = botNum;
	lastTeammateOrigin = origin;

    retTeammate.pos.setAngleNorm(HALF_CIRCLE);
    //if requesting the global coordinates of teammate
    if (origin=='g') {
        //i store in global anyway, so just return the right one
        retTeammate = wmTeammate[botNum-1];
    }
    //else return the coordinates of teammate relative to you
    else {
        retTeammate.vobType = wmTeammate[botNum-1].vobType;
        double dx = wmTeammate[botNum-1].pos.x - r[0].mean.pos.x;
        double dy = wmTeammate[botNum-1].pos.y - r[0].mean.pos.y;
        retTeammate.pos.setVector(vCART,dx,dy);
        retTeammate.posVar = wmTeammate[botNum-1].posVar;
        //
        //INACCURACY ALERT
        //ERIC when converting to local, you should add your own variance again and recalculate posVar
        //
        retTeammate.h = wmTeammate[botNum-1].h;
        retTeammate.hVar = wmTeammate[botNum-1].hVar;
		retTeammate.behavioursVal = wmTeammate[botNum-1].behavioursVal;
		retTeammate.counter = wmTeammate[botNum-1].counter;
        retTeammate.pos.rotate(90-r[0].mean.h);
    }
    return retTeammate;
}

const WMObj& GPS::tmObj(int arrIndex) {
	return wmTeammate[arrIndex];
}

const WMObj& GPS::tmBallObj(int arrIndex) {
	return tmBall[arrIndex];
}

WMObj& GPS::oGoal() {

    //check cache
    if (oGoalAccessId==frameId) {
		//cout << "using cached oGoal" << endl;
		return retOGoal;
	}
	oGoalAccessId = frameId;

    retOGoal.pos.setAngleNorm(HALF_CIRCLE);
    retOGoal.vobType = ownGoal;
    double gx = (FIELD_WIDTH / 2) - r[0].mean.pos.x;
    double gy = (-WALL_THICKNESS) - r[0].mean.pos.y;
    retOGoal.pos.setVector(vCART,gx,gy);
    retOGoal.posVar = r[0].mean.posVar;
    //h and hvar dont apply to the goal
    retOGoal.pos.rotate(90-r[0].mean.h);

    return retOGoal;
}

WMObj& GPS::tGoal() {

    //check cache
    if (tGoalAccessId==frameId) {
		//cout << "using cached tGoal" << endl;
		return retTGoal;
	}

	tGoalAccessId = frameId;

    retTGoal.pos.setAngleNorm(HALF_CIRCLE);
    retTGoal.vobType = targetGoal;
    double gx = (FIELD_WIDTH / 2) - r[0].mean.pos.x;
    double gy = (FIELD_LENGTH + WALL_THICKNESS) - r[0].mean.pos.y;
    retTGoal.pos.setVector(vCART,gx,gy);
    retTGoal.posVar = r[0].mean.posVar;
    //h and hvar dont apply to the goal
    retTGoal.pos.rotate(90-r[0].mean.h);

    return retTGoal;
}


//
// CONSIDERATION
// right now the "location" of the goal is behind the goal line
// so if you are to the side of the goal, you will try and shoot
// to a point within the goal, and the ball will probably bounce
// off the walls beside the opening of the goal..
// Perhaps we should change the "location" of the goal to be
// the mid point of the goal line..
// Or be thinking about the 2 goal posts instead of one point
//


const Vector&
GPS::getOppPos(int oppNum, int context) {
	
	return opponents[oppNum]->getPos(context);
	
}

const MMatrix2&
GPS::getOppCov(int oppNum) {
	
	return opponents[oppNum]->getCov();
	
}

double 
GPS::getOppCovMax(int oppNum) {
    const MMatrix2 mm = opponents[oppNum]->getCov();
    double b = -(mm(0,0) + mm(1,1));
    double c = mm(0,0)*mm(1,1) - mm(0,1)*mm(1,0);
    double posVar;
    if(b*b-4*c < 1E-5) {
        posVar = (-b+0)/2;
        if (b*b-4*c < -1E-5)
            cout << "WARNING: b*b-4*c < 0 (location 1): " << (b*b-4*c) << endl;
    } else {
        posVar = (-b + sqrt(b*b-4*c)) / 2;
    }
    return posVar;
}

#ifdef VISUAL_TRACK_FRIENDS
const Vector&
GPS::getFriendPos(int friendNum, int context) {
	
	return friends[friendNum]->getPos(context);
	
}
#endif // VISUAL_TRACK_FRIENDS

const Vector&
GPS::getClosestOppPos(int context, int *oppNum) {
	
	double minDist = LARGE_VAL;
	int minInd = -1;
	for(int i=0 ; i<NUM_OPPONENTS ; i++) {
		const Vector& v = opponents[i]->getPos(LOCAL);
		if(v.d < minDist) {
			minDist = v.d;
			minInd = i;
		}
	}
	if (oppNum!=NULL) {
		*oppNum = minInd;
	}
	return opponents[minInd]->getPos(context);
	
}



//ERIC!! WOULDNT IT BE BETTER TO DO SOME SORT OF MOD??
// inline function ANGLE_FIX converts the given angle to [0,FULL_CIRCLE (360))
inline void GPS::ANGLE_FIX(double& x) {
    //x=x%(2*M_PI);

    while(x>=FULL_CIRCLE)
        x-=FULL_CIRCLE;
    while(x<0)
        x+=FULL_CIRCLE;
}

// inline function NORMALISE makes the given vector (x,y,z) length 1
inline void GPS::NORMALISE(double& x, double& y, double& z) {
    double l = sqrt(SQUARE(x)+SQUARE(y)+SQUARE(z));
    if (l < 1E-5 && l > -1E-5)
    {
      cout << "WARNING: GPS->NORMALISE is trying to divide by " << l << endl;
      return;
    }
    x /= l;
    y /= l;
    z /= l;
}

// inline function NORMALISE makes the given vector (x,y) length 1
inline void GPS::NORMALISE(double& x, double& y) {
    double l = sqrt(SQUARE(x)+SQUARE(y));
    if (l < 1E-5 && l > -1E-5)
    {
      cout << "WARNING: GPS->NORMALISE is trying to divide by " << l << endl;
      return;
    }
    x /= l;
    y /= l;
}


//WMShareObj share;

//called by shareEnvironment function in UNSWSkills which is called by the strategy files
WMShareObj GPS::getShareInfo () {
    WMShareObj share;

    
    if (	(r[0].mean.pos.x != r[0].mean.pos.x) ||
    		(r[0].mean.pos.y != r[0].mean.pos.y) ||
    		(r[0].mean.h != r[0].mean.h) ||
    		(r[0].mean.posVar != r[0].mean.posVar) ||
    		(r[0].mean.hVar != r[0].mean.hVar) ||
    		(r[0].mean.behavioursVal != r[0].mean.behavioursVal)) {
    	cout << "NaN when trying to share gps info!" << endl;
    	share.posx = share.posy = share.posh = share.posVar = 
    		share.hVar = 0;
    	share.behavioursVal = 0;
    } else {
		share.posx = (float)r[0].mean.pos.x;
		share.posy = (float)r[0].mean.pos.y;
		share.posh = (float)r[0].mean.h;
		share.posVar = (float)r[0].mean.posVar;
		share.hVar = (float)r[0].mean.hVar;
		share.behavioursVal = r[0].mean.behavioursVal;
		/*
		share.ballDist = r[0].mean.ballDist;
		share.playerType = r[0].mean.playerType;
		share.amThirdPlayer = r[0].mean.amThirdPlayer;
		*/
	}
    		
	// Ted : Send ball info to teammates either I can see the ball or I can't see the ball but I am sure the ball is there.
    //       Right? The ball sent is the gps kalman ball even if the robot can see vision ball. However for some reason the
	//       current gps kalman ball is very poor. Derrick is fixing it right now. Come back later. Or use the vision ball
	//       to caluclate the new share.ball ......
	if ((canSee(vobBall) || getBallMaxVar() < get95CF(20)) &&  self().posVar < get95CF(100)) {
        const Vector& b = getBall(GLOBAL);
        share.ballx = b.x;
        share.bally = b.y;
        share.ballvar = getBallMaxVar();
    } else {
		share.ballx = 0;
		share.bally = 0;
        share.ballvar = LARGE_VAL;
	}
    
    for(int i=0 ; i<NUM_OPPONENTS ; i++) {
        opponents[i]->flushUpdate(share.opponentsUpdate[i]);
	}
	
#ifdef IR_LATENCY_TEST
	if(doLatTest) {
		share.latMode = INIT_LAT_WM;
		doLatTest = false;
	} else
		share.latMode = NORMAL_WM;
#endif // IR_LATENCY_TEST
        
    return share;
}

//called by wreceive in UNSWSKills which is called by processWireless in Vision
//which is called by resultWireless in Vision, which is called when a message is received from
//the wireless object
//@param sourceBot is the robot number where the info came from
void GPS::processShareInfo (int sourceBot, const WMShareObj& info) {
    //wmTeammate[sourceBot-1] = info->obj;
	wmTeammate[sourceBot-1].pos.setVector(vCART, (double)info.posx, (double)info.posy);
	wmTeammate[sourceBot-1].h = (double)info.posh;
    wmTeammate[sourceBot-1].posVar = (double)info.posVar;
    wmTeammate[sourceBot-1].hVar = (double)info.hVar;

	wmTeammate[sourceBot-1].behavioursVal = info.behavioursVal;
	
	//bottom stuff is for maverick only and should go away!
	
	
	//at padua i (eric) got rid of these from wireless transmission as they were only used by maverick
	//i replaced them with dummy values so that the code still compiles
	//
#ifdef MAVERICK_WM_STUFF
	wmTeammate[sourceBot-1].ballDist = 400; //info.ballDist;
	wmTeammate[sourceBot-1].playerType = 1; //info.playerType;
	wmTeammate[sourceBot-1].amThirdPlayer = 1; //info.amThirdPlayer;
#endif
	
    //*wmTeammateCov[sourceBot-1] = info->cov;
	wmTeammate[sourceBot-1].counter = TEAMMATE_GPS_COUNTER;

	//store where the teammate thinks the ball is
	tmBall[sourceBot-1].posVar = info.ballvar;
	tmBall[sourceBot-1].pos.setVector(vCART, info.ballx, info.bally);


    if((sourceBot == sbRobotNum) || (info.ballvar < shareBallvar)) {
        shareBallx = info.ballx;
        shareBally = info.bally;
        shareBallvar = info.ballvar;
        sbRobotNum = sourceBot;
    }
	
	for(int i=0 ; i<NUM_OPPONENTS ; i++) {
	    opponents[i]->addUpdates(info.opponentsUpdate[i]);
	}
}


// returns non-zero if it can "see" (ie - using for localisation) the
// given object, according to indexing in visual cortex.
int GPS::canSee(int object) {
    return (view & (1 << object));
}

void GPS::increaseSelfAngleVariance(double amount) {
    for (int i = 0; i < numGaussians; i++) {
        r[i].cov(2,2) += amount;
        r[i].mean.updateVar(&r[i].cov);
    }
}    
